from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers

from ..models import StoredDriver


class StoredDriverSerializer(serializers.HyperlinkedModelSerializer):
    arguments = serializers.SerializerMethodField(
        label=_(message='Arguments')
    )
    description = serializers.SerializerMethodField(
        label=_(message='Description')
    )
    label = serializers.CharField(
        label=_(message='Label'), source='driver_label'
    )
    mime_types = serializers.SerializerMethodField(
        label=_(message='MIME types')
    )
    enabled_default = serializers.SerializerMethodField(
        label=_(message='Enabled default')
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_field': 'pk', 'lookup_url_kwarg': 'stored_driver_id',
                'view_name': 'rest_api:file_metadata_driver-detail'
            }
        }
        fields = (
            'arguments', 'description', 'driver_path', 'enabled_default',
            'id', 'internal_name', 'label', 'mime_types', 'url'
        )
        model = StoredDriver
        read_only_fields = (
            'arguments', 'description', 'driver_path', 'enabled_default',
            'id', 'internal_name', 'label', 'mime_types', 'url'
        )

    def get_arguments(self, instance):
        driver_class = instance.driver_class

        return driver_class.get_argument_values_from_settings_display()

    def get_description(self, instance):
        driver_class = instance.driver_class

        return driver_class.description

    def get_enabled_default(self, instance):
        driver_class = instance.driver_class

        return driver_class.get_enabled_value()

    def get_mime_types(self, instance):
        driver_class = instance.driver_class

        return driver_class.get_mime_type_list_display()



from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.fields import ImageField
from mayan.apps.forms import form_widgets, forms

from .models import SignatureCapture
from .widgets import SignatureCapturesAppWidget


class SignatureCaptureForm(forms.ModelForm):
    class Meta:
        fields = ('data', 'svg', 'text', 'internal_name')
        model = SignatureCapture
        widgets = {
            'data': SignatureCapturesAppWidget(),
            'svg': form_widgets.HiddenInput(
                attrs={
                    'class': 'signature-captures-capture-svg'
                }
            )
        }


class SignatureCaptureDetailForm(forms.DetailForm):
    preview = ImageField(
        image_alt_text=_(message='Asset preview image'), label=_(message='Preview')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['preview'].initial = kwargs['instance']

    class Meta:
        extra_fields = (
            {
                'field': 'date_time_created',
                'widget': form_widgets.DateTimeInput
            },
            {
                'field': 'date_time_edited',
                'widget': form_widgets.DateTimeInput
            }
        )
        fields = ('internal_name', 'text', 'user', 'preview')
        model = SignatureCapture



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link, Separator
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import (
    icon_about, icon_book, icon_documentation, icon_forum,
    icon_knowledge_base, icon_license, icon_mailing_list, icon_object_copy,
    icon_release_notes, icon_setup, icon_source_code, icon_store,
    icon_support, icon_tools
)
from .literals import (
    URL_BOOK, URL_DOCUMENTATION, URL_FORUM, URL_KNOWLEDGE_BASE,
    URL_MAILING_LIST, URL_RELEASE_NOTES, URL_SOURCE_CODE, URL_STORE,
    URL_SUPPORT
)
from .permissions import permission_object_copy


def object_copy_conditional_disable(context, resolved_object):
    # Hidden import.
    from .classes import ModelCopy

    if not resolved_object:
        return False
    else:
        try:
            return ModelCopy.get(
                model=resolved_object._meta.model
            ).test_condition(instance=resolved_object)
        except KeyError:
            return False


link_about = Link(
    icon=icon_about, text=_(message='About this'), view='common:about_view'
)
link_book = Link(
    icon=icon_book, tags='new_window', text=_(message='Get the book'),
    url=URL_BOOK
)
link_documentation = Link(
    icon=icon_documentation, tags='new_window',
    text=_(message='Documentation'), url=URL_DOCUMENTATION
)
link_forum = Link(
    icon=icon_forum, tags='new_window', text=_(message='Forum'),
    url=URL_FORUM
)
link_knowledge_base = Link(
    icon=icon_knowledge_base, tags='new_window',
    text=_(message='Knowledge base'), url=URL_KNOWLEDGE_BASE
)
link_license = Link(
    icon=icon_license, text=_(message='License'), view='common:license_view'
)
link_mailing_list = Link(
    icon=icon_mailing_list, text=_(message='Mailing list'),
    url=URL_MAILING_LIST
)
link_object_copy = Link(
    condition=object_copy_conditional_disable,
    icon=icon_object_copy, kwargs=get_content_type_kwargs_factory(),
    text=_(message='Copy'), permission=permission_object_copy,
    view='common:object_copy'
)
link_release_notes = Link(
    icon=icon_release_notes, tags='new_window',
    text=_(message='Release notes'), url=URL_RELEASE_NOTES
)
link_separator_information = Separator()
link_setup = Link(
    icon=icon_setup, text=_(message='Setup'), view='common:setup_list'
)
link_source_code = Link(
    icon=icon_source_code, tags='new_window', text=_(message='Source code'),
    url=URL_SOURCE_CODE
)
link_store = Link(
    icon=icon_store, tags='new_window', text=_(message='Online store'),
    url=URL_STORE
)
link_support = Link(
    icon=icon_support, tags='new_window', text=_(message='Support'),
    url=URL_SUPPORT
)
link_tools = Link(
    icon=icon_tools, text=_(message='Tools'), view='common:tools_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Smart settings'), name='smart_settings'
)

permission_settings_edit = namespace.add_permission(
    label=_(message='Edit settings'), name='permission_settings_edit'
)
permission_settings_view = namespace.add_permission(
    label=_(message='View settings'), name='permission_settings_view'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import icon_template_sandbox
from .permissions import permission_template_sandbox

link_object_template_sandbox = Link(
    icon=icon_template_sandbox, kwargs=get_content_type_kwargs_factory(
        variable_name='resolved_object'
    ), permission=permission_template_sandbox, text=_(message='Sandbox'),
    view='templating:object_template_sandbox'
)



from mayan.apps.icons.icons import Icon

icon_ajax_refresh = Icon(
    driver_name='fontawesome', symbol='sync'
)

icon_menu_actions = Icon(driver_name='fontawesome', symbol='ellipsis-v')
icon_menu_views = Icon(driver_name='fontawesome', symbol='eye')



from django.dispatch import Signal

signal_perform_upgrade = Signal(use_caching=True)
signal_post_initial_setup = Signal(use_caching=True)
signal_post_upgrade = Signal(use_caching=True)
signal_pre_initial_setup = Signal(use_caching=True)
signal_pre_upgrade = Signal(use_caching=True)
signal_mayan_pre_save = Signal(use_caching=True)



class StoredCredentialBusinessLogicMixin:
    def get_backend_data(self):
        obj = super().get_backend_data()

        if self.backend_path:
            backend_class = self.get_backend_class()

            if hasattr(backend_class, 'post_processing'):
                obj = backend_class.post_processing(obj=obj)

        return obj



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.permissions import permission_settings_view
from mayan.apps.views.generics import SingleObjectListView

from ..classes import FileMetadataDriver
from ..icons import icon_file_metadata, icon_file_metadata_driver_list


class FileMetadataDriverListView(SingleObjectListView):
    view_icon = icon_file_metadata_driver_list
    view_permission = permission_settings_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_file_metadata,
            'no_results_text': _(
                message='File metadata drivers extract embedded information '
                'from document files. File metadata drivers are configure '
                'in code only.'
            ),
            'no_results_title': _(
                message='No file metadata drivers available.'
            ),
            'subtitle': _('File metadata drivers enabled and detected.'),
            'title': _(
                message='File metadata drivers'
            )
        }

    def get_source_queryset(self):
        return FileMetadataDriver.collection.get_all(sorted=True)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0025_auto_20211201_0842')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='indexinstancenode',
            options={
                'ordering': ('value',),
                'verbose_name': 'Index instance node',
                'verbose_name_plural': 'Indexes instances node'
            }
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class SourceCompressedApp(MayanAppConfig):
    app_namespace = 'source_compressed'
    app_url = 'source_compressed'
    has_tests = True
    name = 'mayan.apps.source_compressed'
    verbose_name = _(message='Source compressed')



import hashlib

from django.apps import apps
from django.core import serializers
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from ..literals import GRAPHVIZ_SYMBOL_CONDITIONAL, GRAPHVIZ_SYMBOL_TRIGGER


class WorkflowTransitionBusinessLogicMixin:
    def do_diagram_generate(self, diagram):
        if self.has_condition():
            transition_label = '{} {}'.format(
                self.label, GRAPHVIZ_SYMBOL_CONDITIONAL
            )
        else:
            transition_label = self.label

        if self.trigger_events.exists():
            transition_label = '{} {}'.format(
                transition_label, GRAPHVIZ_SYMBOL_TRIGGER
            )

        edge_kwargs = {
            'head_name': self.destination_state.get_graph_id(),
            'label': '''<
                <table border="3" cellborder="0" color="white">
                    <tr>
                        <td bgcolor="white">{}</td>
                    </tr>
                </table>
            >'''.format(transition_label),
            'tail_name': self.origin_state.get_graph_id()
        }
        diagram.edge(**edge_kwargs)

    def do_execute(
        self, workflow_instance, comment=None, extra_data=None, user=None
    ):
        WorkflowInstanceLogEntry = apps.get_model(
            app_label='document_states', model_name='WorkflowInstanceLogEntry'
        )

        with transaction.atomic():
            workflow_instance.state_active = self.destination_state
            workflow_instance.save()

            workflow_instance_log_entry = WorkflowInstanceLogEntry(
                comment=comment, extra_data=extra_data, transition=self,
                user=user, workflow_instance=workflow_instance
            )
            workflow_instance_log_entry._event_actor = user
            workflow_instance_log_entry.save()

        return workflow_instance_log_entry

    def get_field_display(self):
        field_list = [
            str(field) for field in self.fields.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_field_display.short_description = _(message='Fields')

    def get_form_schema(self, schema, workflow_instance):
        for field in self.fields.all():
            field.get_form_schema(
                schema=schema, workflow_instance=workflow_instance
            )

    def get_hash(self):
        result = hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        )
        for trigger_event in self.trigger_events.all():
            result.update(
                trigger_event.get_hash().encode()
            )

        for field in self.fields.all():
            result.update(
                field.get_hash().encode()
            )

        return result.hexdigest()



from django.urls import re_path

from .api_views import (
    APISignatureCaptureDetailView, APISignatureCapturesImageView,
    APISignatureCaptureListView
)
from .views import (
    SignatureCaptureCreateView, SignatureCaptureDeleteView,
    SignatureCaptureDetailView, SignatureCaptureEditView,
    SignatureCaptureListView
)

urlpatterns_documents = [
    re_path(
        route=r'^documents/(?P<document_id>\d+)/signature_captures/$',
        name='signature_capture_list',
        view=SignatureCaptureListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/signature_captures/create/$',
        name='signature_capture_create',
        view=SignatureCaptureCreateView.as_view()
    )
]

urlpatterns_signature_captures = [
    re_path(
        route=r'^signature_captures/(?P<signature_capture_id>\d+)/delete/$',
        name='signature_capture_delete',
        view=SignatureCaptureDeleteView.as_view()
    ),
    re_path(
        route=r'^signature_captures/(?P<signature_capture_id>\d+)/detail/$',
        name='signature_capture_detail',
        view=SignatureCaptureDetailView.as_view()
    ),
    re_path(
        route=r'^signature_captures/(?P<signature_capture_id>\d+)/edit/$',
        name='signature_capture_edit',
        view=SignatureCaptureEditView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_documents)
urlpatterns.extend(urlpatterns_signature_captures)

api_urls = [
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/signature_captures/$',
        name='signature_capture-list',
        view=APISignatureCaptureListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/signature_captures/(?P<signature_capture_id>[0-9]+)/$',
        name='signature_capture-detail',
        view=APISignatureCaptureDetailView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/signature_captures/(?P<signature_capture_id>[0-9]+)/image/$',
        name='signature_capture-image',
        view=APISignatureCapturesImageView.as_view()
    )
]



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class PlatformsGitlabApp(MayanAppConfig):
    app_namespace = 'platforms_gitlab'
    app_url = 'platforms_gitlab'
    name = 'mayan.apps.platforms_gitlab'
    verbose_name = _(message='Platforms GitLab')



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='User management'), name='user_management'
)

permission_group_create = namespace.add_permission(
    label=_(message='Create new groups'), name='group_create'
)
permission_group_delete = namespace.add_permission(
    label=_(message='Delete existing groups'), name='group_delete'
)
permission_group_edit = namespace.add_permission(
    label=_(message='Edit existing groups'), name='group_edit'
)
permission_group_view = namespace.add_permission(
    label=_(message='View existing groups'), name='group_view'
)
permission_user_create = namespace.add_permission(
    label=_(message='Create new users'), name='user_create'
)
permission_user_delete = namespace.add_permission(
    label=_(message='Delete existing users'), name='user_delete'
)
permission_user_edit = namespace.add_permission(
    label=_(message='Edit existing users'), name='user_edit'
)
permission_user_view = namespace.add_permission(
    label=_(message='View existing users'), name='user_view'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import UserMailer
from .permissions import permission_mailing_profile_use


class MailingProfileSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'mailing_profile_id',
                'view_name': 'rest_api:mailing_profile-detail'
            }
        }
        fields = (
            'backend_data', 'backend_path', 'default', 'enabled', 'id',
            'label', 'url'
        )
        model = UserMailer


class MailingProfileActionSerializer(serializers.Serializer):
    body = serializers.CharField(
        label=_(message='Body'), required=False
    )
    mailing_profile = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the mailing profile to use.'
        ), label=_(message='Mailing profile ID'),
        source_queryset=UserMailer.objects.filter(enabled=True),
        source_permission=permission_mailing_profile_use
    )
    recipient = serializers.CharField(
        label=_(message='Recipient'), required=True
    )
    subject = serializers.CharField(
        label=_(message='Subject'), required=False
    )

    class Meta:
        fields = ('body', 'mailing_profile', 'recipient', 'subject')



import logging

from django.utils.translation import gettext_lazy as _

from mayan.apps.document_states.classes import WorkflowAction

from mayan.apps.document_states.models.workflow_instance_models import (
    WorkflowInstance
)
from mayan.apps.user_management.querysets import get_user_queryset

from .models import Message

logger = logging.getLogger(name=__name__)


class WorkflowActionMessageSend(WorkflowAction):
    form_fields = {
        'group_name_list': {
            'label': _(message='Group name list'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    message='Comma separated list of user group names '
                    'that will receive the message. Can be a static value or '
                    'a template.'
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'role_name_list': {
            'label': _(message='Role name list'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    message='Comma separated list of role labels '
                    'that will receive the message. Can be a static value or '
                    'a template.'
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'username_list': {
            'label': _(message='Username list'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    _(
                        message='Comma separated list of usernames that will '
                        'receive the message. Can be a static value or '
                        'a template.'
                    ),
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'subject': {
            'label': _(message='Subject'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    _(
                        message='Subject of the message to be sent. Can be a '
                        'static value or a template.'
                    )
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        },
        'body': {
            'label': _(message='Body'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    _(
                        message='The actual text to send. Can be a static '
                        'value or a template.'
                    )
                ),
                'model': WorkflowInstance,
                'model_variable': 'workflow_instance',
                'required': True
            }
        }
    }
    label = _(message='Send user message')

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Recipients'), {
                    'fields': (
                        'group_name_list', 'role_name_list', 'username_list'
                    )
                },
            ), (
                _(message='Content'), {
                    'fields': ('subject', 'body')
                }
            )
        )
        return fieldsets

    def execute(self, context):
        final_username_list = []
        queryset_users = get_user_queryset()

        # Group name list.

        group_name_list = self.render_field(
            field_name='group_name_list', context=context
        ) or ''
        group_name_list = group_name_list.split(',')

        if group_name_list:
            queyset_group_users = queryset_users.filter(
                groups__name__in=group_name_list
            )
            final_username_list.extend(
                queyset_group_users.values_list('username', flat=True)
            )

        # Role label list.

        role_label_list = self.render_field(
            field_name='role_label_list', context=context
        ) or ''
        role_label_list = role_label_list.split(',')

        if role_label_list:
            queyset_role_users = queryset_users.filter(
                groups__roles__label__in=role_label_list
            )
            final_username_list.extend(
                queyset_role_users.values_list('username', flat=True)
            )

        # Username list.

        username_list = self.render_field(
            field_name='username_list', context=context
        ) or ''
        username_list = username_list.split(',')
        if username_list:
            final_username_list.extend(username_list)

        subject = self.render_field(
            field_name='subject', context=context
        ) or ''

        body = self.render_field(
            field_name='body', context=context
        ) or ''

        queryset_users_to_message = queryset_users.filter(
            username__in=final_username_list
        )

        for user in queryset_users_to_message:
            Message.objects.create(body=body, subject=subject, user=user)



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import mayan.apps.views.model_mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='UserViewMode',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'namespace', models.CharField(
                        db_index=True, max_length=200,
                        verbose_name='Namespace'
                    )
                ),
                (
                    'name', models.CharField(
                        db_index=True,
                        help_text='Full name of the view including the '
                        'namespace.', max_length=200, verbose_name='Name'
                    )
                ),
                (
                    'value', models.CharField(
                        db_index=True,
                        help_text='Stored value used to identify the '
                        'display mode of the view.', max_length=5,
                        verbose_name='Value'
                    )
                ),
                (
                    'user', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='view_modes',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                )
            ],
            options={
                'ordering': ('user__username', 'name'),
                'unique_together': {
                    ('user', 'name')
                },
                'verbose_name': 'User view mode',
                'verbose_name_plural': 'User view modes'
            },
            bases=(
                mayan.apps.views.model_mixins.ModelMixinUserViewModeBusinessLogic,
                models.Model
            )
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0034_workflowtransitionfield_default')
    ]

    operations = [
        migrations.AddField(
            field=models.BooleanField(
                default=False, help_text='The state at which the workflow '
                'will stop. Only one state can be the final state.',
                verbose_name='Final'
            ),
            model_name='workflowstate', name='final'
        )
    ]



from django.utils.translation import gettext_lazy as _

RELATIONSHIP_NO = 'no'
RELATIONSHIP_YES = 'yes'
RELATIONSHIP_CHOICES = (
    (RELATIONSHIP_NO, _(message='No')),
    (RELATIONSHIP_YES, _(message='Yes')),
)



from mayan.apps.acls.classes import ModelPermission
from mayan.apps.events.classes import EventModelRegistry

from .exceptions import MailerError


class ModelMailingAction:
    _registry = {}
    arguments = ('manager_name', 'permission',)
    as_attachment = False
    name = None

    @classmethod
    def get_action_for_model(cls, model, action_name):
        try:
            model_actions = cls._registry[model]
        except KeyError:
            raise MailerError(
                'Model `{}` is not registered for emailing.'.format(model)
            )
        else:
            try:
                return model_actions[action_name]
            except KeyError:
                raise MailerError(
                    'Model `{}` is not registered for emailing action '
                    '`{}`.'.format(model, action_name)
                )

    def __init__(self, model, **kwargs):
        self.kwargs = {}

        arguments = self.get_arguments()

        for argument in arguments:
            try:
                value = kwargs.pop(argument)
            except KeyError:
                raise MailerError(
                    'Error registering mailer action `{}` for model `{}`. '
                    'Missing argument `{}`.'.format(
                        self.name, model, argument
                    )
                )

            self.kwargs[argument] = value

        if kwargs:
            raise MailerError(
                'Error registering mailer action `{}` for model `{}`. '
                'Too many keyword arguments `{}`.'.format(
                    self.name, model, kwargs
                )
            )

        self.__class__._registry.setdefault(
            model, {}
        )

        self.__class__._registry[model][self.name] = self

        EventModelRegistry.register(model=model)

        ModelPermission.register(
            model=model, permissions=(
                self.kwargs['permission'],
            )
        )

    def get_arguments(self):
        return self.arguments


class ModelMailingActionAttachment(ModelMailingAction):
    as_attachment = True
    name = 'attachment'

    def get_arguments(self):
        arguments = super().get_arguments()

        arguments += (
            (
                'content_function_dotted_path',
                'mime_type_function_dotted_path'
            )
        )

        return arguments


class ModelMailingActionLink(ModelMailingAction):
    name = 'link'



from django.contrib import messages
from django.templatetags.static import static
from django.utils.translation import gettext_lazy as _
from django.views.generic import RedirectView

from stronghold.views import StrongholdPublicMixin

from mayan.apps.views.generics import ConfirmView, SimpleView
from mayan.apps.views.view_mixins import (
    ExternalContentTypeObjectViewMixin, ObjectNameViewMixin
)

from .classes import ModelCopy
from .forms import LicenseForm
from .icons import (
    icon_about, icon_home, icon_license, icon_object_copy, icon_setup,
    icon_tools
)
from .menus import menu_tools, menu_setup
from .permissions import permission_object_copy
from .settings import setting_home_view


class AboutView(SimpleView):
    extra_context = {
        'title': _(message='About')
    }
    template_name = 'appearance/about.html'
    view_icon = icon_about


class FaviconRedirectView(RedirectView):
    permanent = True

    def get_redirect_url(self, *args, **kwargs):
        """
        Hide the static tag import to avoid errors with static file
        processors.
        """
        return static(path='appearance/images/favicon.ico')


class HomeView(SimpleView):
    extra_context = {
        'title': _(message='Home')
    }
    template_name = 'appearance/home.html'
    view_icon = icon_home


class LicenseView(SimpleView):
    extra_context = {
        'form': LicenseForm(),
        'read_only': True,
        'title': _(message='License')
    }
    template_name = 'appearance/form_container.html'
    view_icon = icon_license


class ObjectCopyView(
    ExternalContentTypeObjectViewMixin, ObjectNameViewMixin, ConfirmView
):
    external_object_permission = permission_object_copy
    view_icon = icon_object_copy

    def get_extra_context(self):
        model_copy = ModelCopy.get(model=self.external_object._meta.model)
        context = {
            'object': self.external_object,
            'subtitle': _(message='Fields to be copied: %s') % ', '.join(
                sorted(
                    map(
                        str, model_copy.get_fields_verbose_names()
                    )
                )
            )
        }

        context['title'] = _(
            message='Make a copy of %(object_name)s "%(object)s"?'
        ) % {
            'object_name': self.get_object_name(context=context),
            'object': self.external_object
        }

        return context

    def view_action(self):
        self.external_object.copy_instance()
        messages.success(
            message=_(message='Object copied successfully.'),
            request=self.request
        )


class RootView(StrongholdPublicMixin, SimpleView):
    extra_context = {'home_view': setting_home_view.value}
    template_name = 'appearance/root.html'


class SetupListView(SimpleView):
    template_name = 'appearance/list_horizontal.html'
    view_icon = icon_setup

    def get_extra_context(self, **kwargs):
        return {
            'no_results_icon': icon_setup,
            'no_results_text': _(
                message='No results here means that don\'t have the required '
                'permissions to perform administrative task.'
            ),
            'no_results_title': _(message='No setup options available.'),
            'resolved_links': menu_setup.resolve(
                request=self.request, sort_results=True
            ),
            'subtitle': _(
                message='Here you can configure all aspects of the system.'
            ),
            'title': _(message='Setup items')
        }


class ToolsListView(SimpleView):
    template_name = 'appearance/list_horizontal.html'
    view_icon = icon_tools

    def get_extra_context(self):
        return {
            'resolved_links': menu_tools.resolve(
                request=self.request, sort_results=True
            ),
            'subtitle': _(
                message='These modules are used to do system maintenance.'
            ),
            'title': _(message='Tools')
        }



import functools
import logging
from pathlib import Path

import whoosh
from whoosh import qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.index import EmptyIndexError
from whoosh.qparser import MultifieldParser, OrGroup
from whoosh.query import Every
from whoosh.writing import BufferedWriter

from django.conf import settings

from mayan.apps.common.utils import any_to_bool
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.storage.utils import TemporaryDirectory

from ...exceptions import (
    DynamicSearchBackendException, DynamicSearchRetry,
    DynamicSearchValueTransformationError
)
from ...search_backends import SearchBackend
from ...search_fields import SearchFieldVirtualAllFields
from ...search_models import SearchModel

from .literals import (
    DJANGO_TO_WHOOSH_FIELD_MAP, TEXT_LOCK_INSTANCE_DEINDEX,
    TEXT_LOCK_INSTANCE_INDEX, WHOOSH_INDEX_DIRECTORY_NAME
)

logger = logging.getLogger(name=__name__)


class WhooshSearchBackend(SearchBackend):
    _local_attribute_backend_temporary_directory = None
    feature_reindex = True
    field_type_mapping = DJANGO_TO_WHOOSH_FIELD_MAP

    def __init__(
        self, index_path=None, writer_limitmb=128, writer_multisegment=False,
        writer_procs=1, **kwargs
    ):
        super().__init__(**kwargs)

        if self._test_mode:
            if not self.__class__._local_attribute_backend_temporary_directory:
                self.__class__._local_attribute_backend_temporary_directory = TemporaryDirectory()
            index_path = self._local_attribute_backend_temporary_directory.name

        self.index_path = Path(
            index_path or Path(
                settings.MEDIA_ROOT, WHOOSH_INDEX_DIRECTORY_NAME
            )
        )

        if writer_limitmb:
            writer_limitmb = int(writer_limitmb)

        if writer_multisegment:
            writer_multisegment = any_to_bool(value=writer_multisegment)

        if writer_procs:
            writer_procs = int(writer_procs)

        self.writer_kwargs = {
            'limitmb': writer_limitmb, 'multisegment': writer_multisegment,
            'procs': writer_procs
        }

    def _clear_search_model_index(self, search_model):
        schema = self._get_search_model_schema(search_model=search_model)

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            # Clear the model index.
            self._get_storage().create_index(
                indexname=search_model.full_name, schema=schema
            )

    def _do_query_resolve(self, index, query):
        with index.searcher() as searcher:
            results = searcher.search(limit=None, q=query)
            logger.debug('results: %s', results)

            for result in results:
                yield int(
                    result['id']
                )

    def _get_or_create_index(self, search_model):
        storage = self._get_storage()
        schema = self._get_search_model_schema(search_model=search_model)

        try:
            # Explicitly specify the schema. Allows using existing index
            # when the schema changes.
            index = storage.open_index(
                indexname=search_model.full_name, schema=schema
            )
        except EmptyIndexError:
            index = storage.create_index(
                indexname=search_model.full_name, schema=schema
            )

        return index

    @functools.cache
    def _get_search_model_schema(self, search_model):
        field_map = self.get_resolved_field_type_map(
            search_model=search_model
        )
        schema_kwargs = {
            key: value['field'] for key, value in field_map.items()
        }

        return whoosh.fields.Schema(**schema_kwargs)

    def _get_status(self):
        result = []

        title = 'Whoosh search model indexing status'
        result.append(title)
        result.append(
            len(title) * '='
        )

        for search_model in SearchModel.all():
            index = self._get_or_create_index(search_model=search_model)

            with index.searcher() as searcher:
                search_results = searcher.search(
                    q=Every('id')
                )

                result.append(
                    '{}: {}'.format(
                        search_model.label, search_results.estimated_length()
                    )
                )

        return '\n'.join(result)

    def _get_storage(self):
        return FileStorage(path=self.index_path)

    def _get_writer(self, search_model):
        index = self._get_or_create_index(search_model=search_model)

        return index.writer(**self.writer_kwargs)

    def _initialize(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            self.index_path.mkdir(exist_ok=True)

    def _search(
        self, search_field, query_type, value, is_quoted_value=False,
        is_raw_value=False
    ):
        self.do_query_type_verify(
            query_type=query_type, search_field=search_field
        )

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            index = self._get_or_create_index(
                search_model=search_field.search_model
            )

            if isinstance(search_field, SearchFieldVirtualAllFields):
                parser = MultifieldParser(
                    [
                        search_field.field_name for search_field in search_field.field_composition
                    ], schema=index.schema, group=OrGroup
                )
                search_field_queries = []

                for search_field in search_field.field_composition:
                    try:
                        search_string = query_type.resolve_for_backend(
                            is_quoted_value=is_quoted_value,
                            is_raw_value=is_raw_value, search_backend=self,
                            search_field=search_field, value=value,
                            extra_kwargs={
                                'parser': parser
                            }
                        )
                    except DynamicSearchValueTransformationError:
                        """Skip the search field."""
                    else:
                        if search_string is not None:
                            search_field_queries.append(search_string)

                query = parser.parse(
                    ' '.join(search_field_queries)
                )
            else:
                parser = qparser.QueryParser(
                    fieldname=search_field.field_name, schema=index.schema
                )

                try:
                    search_string = query_type.resolve_for_backend(
                        is_quoted_value=is_quoted_value,
                        is_raw_value=is_raw_value, search_backend=self,
                        search_field=search_field, value=value,
                        extra_kwargs={
                            'parser': parser
                        }
                    )
                except DynamicSearchValueTransformationError:
                    return ()
                else:
                    if search_string is None:
                        return ()

                logger.debug('search_string: %s', search_string)

                query = parser.parse(text=search_string)

            return self._do_query_resolve(index=index, query=query)
        else:
            return ()

    def _update_mappings(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            for search_model in search_models:
                self._get_or_create_index(search_model=search_model)

    def deindex_instance(self, instance):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_DEINDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                index = self._get_or_create_index(search_model=search_model)

                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    with index.writer(**self.writer_kwargs) as writer:
                        writer.delete_by_term(
                            'id', str(instance.pk)
                        )
            finally:
                lock.release()

    def do_native_type_conversion(self, value):
        if isinstance(value, (list, tuple)):
            return ' '.join(value)
        else:
            return value

    def index_instance(self, instance, exclude_model=None, exclude_kwargs=None):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_INDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    with self._get_writer(search_model=search_model) as writer:
                        try:
                            writer.delete_by_term(
                                'id', str(instance.pk)
                            )
                        except Exception as exception:
                            # The parenthesis is used to define a multi
                            # line error message not a translatable string.
                            error_text = (
                                'Unexpected exception while '
                                'deleting search object id: {id}, '
                                'search model: {search_model}, '
                                'raw data: {raw_data}, '
                                'field map: {field_map}; '
                                '{exception}'
                            ).format(
                                exception=exception,
                                field_map=self.get_resolved_field_type_map(
                                    search_model=search_model
                                ), id=instance.pk,
                                raw_data=instance.__dict__,
                                search_model=search_model.full_name
                            )

                            logger.error(error_text, exc_info=True)
                            raise DynamicSearchBackendException(
                                error_text
                            ) from exception
                        else:
                            kwargs = search_model.populate(
                                search_backend=self, instance=instance,
                                exclude_model=exclude_model,
                                exclude_kwargs=exclude_kwargs
                            )

                            try:
                                writer.add_document(**kwargs)
                            except Exception as exception:
                                # The parenthesis is used to define a multi
                                # line error message not a translatable
                                # string.
                                error_text = (
                                    'Unexpected exception while '
                                    'indexing search object id: {id}, '
                                    'search model: {search_model}, '
                                    'index data: {index_data}, '
                                    'raw data: {raw_data}, '
                                    'field map: {field_map}; '
                                    '{exception}'
                                ).format(
                                    exception=exception,
                                    field_map=self.get_resolved_field_type_map(
                                        search_model=search_model
                                    ), id=instance.pk, index_data=kwargs,
                                    raw_data=instance.__dict__,
                                    search_model=search_model.full_name
                                )

                                logger.error(error_text, exc_info=True)
                                raise DynamicSearchBackendException(
                                    error_text
                                ) from exception

            except whoosh.index.LockError:
                raise DynamicSearchRetry
            finally:
                lock.release()

    def index_instances(self, search_model, id_list):
        queryset = search_model.get_queryset()
        queryset = queryset.filter(pk__in=id_list)

        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_INDEX
            )
        except LockError:
            raise
        else:
            try:
                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    index = self._get_or_create_index(search_model=search_model)

                    writer = BufferedWriter(index=index)
                    for instance in queryset:
                        kwargs = search_model.populate(
                            search_backend=self, instance=instance
                        )

                        try:
                            writer.update_document(**kwargs)
                        except Exception as exception:
                            # The parenthesis is used to define a multi
                            # line error message not a translatable string.
                            error_text = (
                                'Unexpected exception while '
                                'indexing search model: {search_model}, '
                                'id_list: {id_list}',
                                'index data: {index_data}, '
                                'raw data: {raw_data}, '
                                'field map: {field_map}; '
                                '{exception}'
                            ).format(
                                exception=exception,
                                field_map=self.get_resolved_field_type_map(
                                    search_model=search_model
                                ), id_list=id_list, index_data=kwargs,
                                raw_data=instance.__dict__,
                                search_model=search_model.full_name
                            )

                            logger.error(error_text, exc_info=True)
                            raise DynamicSearchBackendException(
                                error_text
                            ) from exception
                    writer.close()
            except whoosh.index.LockError:
                raise DynamicSearchRetry
            finally:
                lock.release()

    def reset(self, search_model=None):
        self.tear_down(search_model=search_model)
        self._update_mappings(search_model=search_model)

    def tear_down(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            for search_model in search_models:
                self._clear_search_model_index(search_model=search_model)

    def test_mode_stop(self):
        self.__class__._local_attribute_backend_temporary_directory.cleanup()



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave

from ..events import (
    event_workflow_instance_created, event_workflow_instance_transitioned
)
from ..managers import ValidWorkflowInstanceManager

from .workflow_instance_model_mixins import (
    WorkflowInstanceBusinessLogicMixin,
    WorkflowInstanceLogEntryBusinessLogicMixin
)
from .workflow_models import Workflow
from .workflow_state_models import WorkflowState
from .workflow_transition_field_models import WorkflowTransitionField

__all__ = ('WorkflowInstance', 'WorkflowInstanceLogEntry')


class WorkflowInstance(
    ExtraDataModelMixin, WorkflowInstanceBusinessLogicMixin, models.Model
):
    _ordering_fields = ('datetime',)

    workflow = models.ForeignKey(
        help_text=_('The workflow template.'), on_delete=models.CASCADE,
        related_name='instances', to=Workflow,
        verbose_name=_(message='Workflow')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True,
        help_text=_('Workflow instance creation date time.'),
        verbose_name=_(message='Datetime')
    )
    document = models.ForeignKey(
        help_text=_(
            'The document to which the workflow instance is attached.'
        ), on_delete=models.CASCADE, related_name='workflows', to=Document,
        verbose_name=_(message='Document')
    )
    context = models.TextField(
        blank=True, verbose_name=_(message='Context')
    )
    state_active = models.ForeignKey(
        help_text=_('The currently active state of the workflow instance.'),
        on_delete=models.CASCADE, related_name='workflow_instances',
        to=WorkflowState, verbose_name=_(message='Active state')
    )

    objects = models.Manager()
    valid = ValidWorkflowInstanceManager()

    class Meta:
        ordering = ('workflow',)
        unique_together = ('document', 'workflow')
        verbose_name = _(message='Workflow instance')
        verbose_name_plural = _(message='Workflow instances')

    def __str__(self):
        return str(self.workflow)

    def get_absolute_url(self):
        return reverse(
            viewname='document_states:workflow_instance_detail', kwargs={
                'workflow_instance_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'document',
            'event': event_workflow_instance_created,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        created = not self.pk

        if created:
            self.state_active = self.workflow.get_state_initial()

        super().save(*args, **kwargs)


class WorkflowInstanceLogEntry(
    WorkflowInstanceLogEntryBusinessLogicMixin, models.Model
):
    """
    Fields:
    * user - The user who last transitioned the document from a state to the
    Actual State.
    * datetime - Date Time - The date and time when the last user transitioned
    the document state to the Actual state.
    """
    _ordering_fields = ('comment', 'datetime')

    workflow_instance = models.ForeignKey(
        on_delete=models.CASCADE, related_name='log_entries',
        to=WorkflowInstance, verbose_name=_(message='Workflow instance')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_(message='Datetime')
    )
    transition = models.ForeignKey(
        on_delete=models.CASCADE, to='WorkflowTransition',
        verbose_name=_(message='Transition')
    )
    user = models.ForeignKey(
        blank=True, null=True, on_delete=models.CASCADE,
        to=settings.AUTH_USER_MODEL, verbose_name=_(message='User')
    )
    comment = models.TextField(
        blank=True, verbose_name=_(message='Comment')
    )
    extra_data = models.TextField(
        blank=True, verbose_name=_(message='Extra data')
    )

    class Meta:
        ordering = ('datetime',)
        verbose_name = _(message='Workflow instance log entry')
        verbose_name_plural = _(message='Workflow instance log entries')

    def __str__(self):
        return str(self.transition)

    def clean(self):
        queryset = self.workflow_instance.get_queryset_valid_transitions(
            user=self.user
        )
        if not queryset.filter(pk=self.transition.pk).exists():
            raise ValidationError(
                message=_(message='Not a valid transition choice.')
            )

        extra_data = self.loads()

        for field_name, value in extra_data.items():
            try:
                field = self.transition.fields.get(name=field_name)
            except WorkflowTransitionField.DoesNotExist:
                """
                Allow extra data keys that do not match to a know field
                to allow backward compatibility. This may change in the
                future.
                Possible deprecation.
                """
            else:
                field.validate_value(value=value, workflow_instance=self)

    def save(self, *args, **kwargs):
        result = super().save(*args, **kwargs)

        actor = getattr(self, '_event_actor', None)
        event_workflow_instance_transitioned.commit(
            action_object=self.workflow_instance.document, actor=actor,
            target=self.workflow_instance
        )

        self.transition.origin_state.do_active_unset(log_entry=self)
        self.transition.destination_state.do_active_set(log_entry=self)

        return result



from ..models import Notification
from ..serializers.notification_serializers import NotificationSerializer


class APIViewMixinNotification:
    serializer_class = NotificationSerializer

    def get_source_queryset(self):
        if self.request.user.is_authenticated:
            queryset = Notification.objects.filter(user=self.request.user)
        else:
            queryset = Notification.objects.none()

        return queryset



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import (
    GoogleFontDependency, JavaScriptDependency
)

GoogleFontDependency(
    label=_(message='Lato font'), module=__name__, name='lato',
    url='https://fonts.googleapis.com/css?family=Lato:400,700,400italic'
)
JavaScriptDependency(
    label=_(message='Bootstrap'), module=__name__, name='bootstrap',
    version_string='=3.4.1'
)
JavaScriptDependency(
    label=_(message='Bootswatch'), module=__name__, name='bootswatch',
    replace_list=[
        {
            'filename_pattern': 'bootstrap.*.css',
            'content_patterns': [
                {
                    'search': '"https://fonts.googleapis.com/css?family=Lato:400,700,400italic"',
                    'replace': '../../../google_fonts/lato/import.css',
                }
            ]
        }
    ], version_string='=3.4.1'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster
from mayan.settings.literals import (
    GUNICORN_LIMIT_REQUEST_LINE, GUNICORN_MAX_REQUESTS,
    GUNICORN_REQUESTS_JITTER, GUNICORN_TIMEOUT, GUNICORN_WORKER_CLASS,
    GUNICORN_WORKERS
)

from .literals import (
    DEFAULT_SETTINGS_MODULE, DEFAULT_PLATFORMS_CLIENT_BACKEND_ARGUMENTS,
    DEFAULT_PLATFORMS_CLIENT_BACKEND_ENABLED
)


setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Platform'), name='platform'
)

setting_client_backend_enabled = setting_namespace.do_setting_add(
    default=DEFAULT_PLATFORMS_CLIENT_BACKEND_ENABLED,
    global_name='PLATFORMS_CLIENT_BACKEND_ENABLED', help_text=_(
        message='List of client backends to launch after startup. Use full dotted '
        'path to the client backend classes.'
    )
)
setting_client_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_PLATFORMS_CLIENT_BACKEND_ARGUMENTS,
    global_name='PLATFORMS_CLIENT_BACKEND_ARGUMENTS', help_text=_(
        message='Arguments for the client backends. Use the client backend dotted '
        'path as the dictionary key for the arguments in dictionary format.'
    )
)

# Supervisord

setting_namespace.do_setting_add(
    default=DEFAULT_SETTINGS_MODULE,
    global_name='MAYAN_SETTINGS_MODULE', help_text=_(
        message='Load an alternate settings file.'
    )
)

# Gunicorn

setting_namespace.do_setting_add(
    default=GUNICORN_WORKERS,
    global_name='MAYAN_GUNICORN_WORKERS', help_text=_(
        message='Allows setting Gunicorn\'s `workers` value.'
    )
)
setting_namespace.do_setting_add(
    default=GUNICORN_WORKER_CLASS,
    global_name='MAYAN_GUNICORN_WORKER_CLASS', help_text=_(
        message='Allows setting Gunicorn\'s `worker_class` value.'

    )
)
setting_namespace.do_setting_add(
    default=GUNICORN_TIMEOUT,
    global_name='MAYAN_GUNICORN_TIMEOUT', help_text=_(
        message='Allows setting Gunicorn\'s `timeout` value.'
    )
)
setting_namespace.do_setting_add(
    default=GUNICORN_REQUESTS_JITTER,
    global_name='MAYAN_GUNICORN_REQUESTS_JITTER', help_text=_(
        message='Allows setting Gunicorn\'s `max_requests_jitter` value.'
    )
)
setting_namespace.do_setting_add(
    default=GUNICORN_MAX_REQUESTS,
    global_name='MAYAN_GUNICORN_MAX_REQUESTS', help_text=_(
        message='Allows setting Gunicorn\'s `max_requests` value.'
    )
)
setting_namespace.do_setting_add(
    default=GUNICORN_LIMIT_REQUEST_LINE,
    global_name='MAYAN_GUNICORN_LIMIT_REQUEST_LINE', help_text=_(
        message='Allows setting Gunicorn\'s `limit_request_line` value.'
    )
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Signature captures'), name='signature_captures'
)

permission_signature_capture_create = namespace.add_permission(
    label=_(message='Create signature captures'),
    name='signature_capture_create'
)
permission_signature_capture_delete = namespace.add_permission(
    label=_(message='Delete signature captures'),
    name='signature_capture_delete'
)
permission_signature_capture_edit = namespace.add_permission(
    label=_(message='Edit signature captures'),
    name='signature_capture_edit'
)
permission_signature_capture_view = namespace.add_permission(
    label=_(message='View signature captures'),
    name='signature_capture_view'
)



from mayan.apps.acls.models import AccessControlList

from ..models.metadata_type_models import MetadataType
from ..permissions import (
    permission_metadata_type_edit, permission_metadata_type_view
)


class MetadataTypeFilterAPIMixin:
    metadata_type_permission_map = {
        'DELETE': permission_metadata_type_edit,
        'GET': permission_metadata_type_view,
        'PATCH': permission_metadata_type_edit,
        'PUT': permission_metadata_type_edit
    }

    def get_metadata_type_queryset(self):
        permission = self.metadata_type_permission_map.get(
            self.request.method, None
        )

        queryset = MetadataType.objects.all()
        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return queryset

    def get_source_queryset(self):
        return self.get_external_object().metadata.filter(
            metadata_type__in=self.get_metadata_type_queryset()
        )



import re

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.databases.model_mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('documents', '0080_populate_file_size')
    ]

    operations = [
        migrations.CreateModel(
            name='SignatureCapture',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'data', models.TextField(
                        blank=True, help_text='Data representing the '
                        'handwritten signature.',
                        verbose_name='Signature capture data'
                    )
                ),
                (
                    'svg', models.TextField(
                        blank=True, help_text='Vector representation of '
                        'the handwritten signature.',
                        verbose_name='SVG signature capture data'
                    )
                ),
                (
                    'text', models.CharField(
                        help_text='Print version of the captured signature.',
                        max_length=224, verbose_name='Text'
                    )
                ),
                (
                    'date_time_created', models.DateTimeField(
                        auto_now_add=True, db_index=True,
                        verbose_name='Date and time created'
                    )
                ),
                (
                    'date_time_edited', models.DateTimeField(
                        auto_now=True, db_index=True,
                        verbose_name='Date and time edited'
                    )
                ),
                (
                    'internal_name', models.CharField(
                        db_index=True, help_text='This value will be '
                        'used when referencing this signature capture '
                        'in relationship to the document. Can only '
                        'contain letters, numbers, and underscores.',
                        max_length=255, validators=[
                            django.core.validators.RegexValidator(
                                re.compile('^[a-zA-Z0-9_]+\\Z'), "Enter "
                                "a valid 'internal name' consisting of "
                                "letters, numbers, and underscores.",
                                'invalid'
                            )
                        ], verbose_name='Internal name'
                    )
                ),
                (
                    'document', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='signature_captures',
                        to='documents.document', verbose_name='Document'
                    )
                ),
                (
                    'user', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='signature_captures',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                )
            ],
            options={
                'verbose_name': 'Signature capture',
                'verbose_name_plural': 'Signature captures',
                'ordering': ('-date_time_created',),
                'unique_together': {
                    ('document', 'internal_name')
                }
            },
            bases=(
                mayan.apps.databases.model_mixins.ExtraDataModelMixin,
                models.Model
            )
        )
    ]



import hashlib

from django.core import serializers
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.common.utils import comma_splitter
from mayan.apps.forms.literals import EMPTY_LABEL
from mayan.apps.templating.template_backends import Template

from ..literals import FIELD_TYPE_MAPPING, WIDGET_CLASS_MAPPING


class WorkflowTransitionFieldBusinessLogicMixin:
    def get_default_value(self, workflow_instance):
        template = Template(template_string=self.default)
        context = {'workflow_instance': workflow_instance}

        return template.render(context=context)

    def get_form_schema(self, schema, workflow_instance):
        schema['fields'][self.name] = {
            'class': FIELD_TYPE_MAPPING[self.field_type],
            'help_text': self.help_text,
            'label': self.label,
            'required': self.required
        }

        schema_field = schema['fields'][self.name]

        if self.widget:
            schema['widgets'][self.name] = {
                'class': WIDGET_CLASS_MAPPING[self.widget],
                'kwargs': self.get_widget_kwargs()
            }

        if self.default:
            default_value = self.get_default_value(
                workflow_instance=workflow_instance
            )
            schema_field['default'] = default_value

        if self.lookup:
            try:
                lookup_choices = self.get_lookup_values(
                    workflow_instance=workflow_instance
                )

                field_choices = list(
                    zip(lookup_choices, lookup_choices)
                )

                if not self.required:
                    field_choices.insert(
                        0, ('', EMPTY_LABEL)
                    )

                schema_field['class'] = 'django.forms.fields.ChoiceField'
                schema_field['kwargs'] = {'choices': field_choices}

                schema['widgets'][self.name] = {
                    'class': 'django.forms.fields.Select',
                    'kwargs': {
                        'attrs': {'class': 'select2'}
                    }
                }
            except Exception as exception:
                schema_field['class'] = 'django.forms.fields.CharField'
                schema_field['kwargs'] = {
                    'initial': _(
                        message='Lookup value error: %s'
                    ) % exception
                }
                schema['widgets'][self.name] = {
                    'class': 'django.forms.fields.TextInput',
                    'kwargs': {
                        'attrs': {'readonly': 'readonly'}
                    }
                }

    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def get_lookup_values(self, workflow_instance):
        template = Template(
            context_entry_name_list=('groups', 'users'),
            template_string=self.lookup
        )
        return comma_splitter(
            template.render(
                context={'workflow_instance': workflow_instance}
            )
        )

    def get_widget_kwargs(self):
        return yaml_load(
            stream=self.widget_kwargs or '{}'
        )

    def has_default(self):
        if self.default:
            return True
        else:
            return False

    has_default.short_description = _(message='Has a default?')

    def has_lookup(self):
        if self.lookup:
            return True
        else:
            return False

    has_lookup.short_description = _(message='Has a lookup?')

    def validate_value(self, value, workflow_instance):
        if self.lookup:
            lookup_options = self.get_lookup_values(
                workflow_instance=workflow_instance
            )

            if value and value not in lookup_options:
                raise ValidationError(
                    message=_(
                        message='Value is not one of the provided options.'
                    )
                )



class TemplatingError(Exception):
    """
    Base exception for all templating app exceptions.
    """


class DangerousTagError(TemplatingError):
    """
    Raised when a template tries to execute a dangerous tag.
    """



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_document_file_metadata_submit_multiple,
    icon_document_file_metadata_submit_single,
    icon_document_type_file_metadata_driver_configuration_edit,
    icon_document_type_file_metadata_driver_configuration_list,
    icon_document_type_file_metadata_submit, icon_file_metadata,
    icon_file_metadata_driver_list
)
from .permissions import (
    permission_document_type_file_metadata_setup,
    permission_file_metadata_submit, permission_file_metadata_view
)

# Document file

link_document_file_metadata_driver_list = Link(
    icon=icon_file_metadata,
    kwargs={'document_file_id': 'resolved_object.id'},
    permission=permission_file_metadata_view, text=_(message='File metadata'),
    view='file_metadata:document_file_metadata_driver_list'
)
link_document_file_metadata_driver_attribute_list = Link(
    icon=icon_file_metadata,
    kwargs={'document_file_driver_id': 'resolved_object.id'},
    permission=permission_file_metadata_view, text=_(message='Attributes'),
    view='file_metadata:document_file_metadata_driver_attribute_list'
)
link_document_file_metadata_submit_multiple = Link(
    icon=icon_document_file_metadata_submit_multiple,
    text=_(message='Submit for file metadata'),
    view='file_metadata:document_file_metadata_multiple_submit'
)
link_document_file_metadata_submit_single = Link(
    icon=icon_document_file_metadata_submit_single,
    kwargs={'document_file_id': 'resolved_object.id'},
    permission=permission_file_metadata_submit,
    text=_(message='Submit for file metadata'),
    view='file_metadata:document_file_metadata_single_submit'
)

# Document type

link_document_type_file_metadata_driver_configuration_edit = Link(
    icon=icon_document_type_file_metadata_driver_configuration_edit,
    kwargs={
        'document_type_id': 'document_type.id',
        'stored_driver_id': 'resolved_object.stored_driver_id'
    }, permission=permission_document_type_file_metadata_setup,
    text=_(message='Edit'),
    view='file_metadata:document_type_file_metadata_driver_configuration_edit'
)
link_document_type_file_metadata_driver_configuration_list = Link(
    icon=icon_document_type_file_metadata_driver_configuration_list,
    kwargs={'document_type_id': 'resolved_object.id'},
    permission=permission_document_type_file_metadata_setup,
    text=_(message='File metadata driver setup'),
    view='file_metadata:document_type_file_metadata_driver_configuration_list'
)
link_document_type_file_metadata_submit = Link(
    icon=icon_document_type_file_metadata_submit,
    permission=permission_file_metadata_submit,
    text=_(message='File metadata processing per type'),
    view='file_metadata:document_type_file_metadata_submit'
)

# Tools

link_file_metadata_driver_list = Link(
    icon=icon_file_metadata_driver_list,
    permission=permission_file_metadata_view,
    text=_(message='File metadata drivers'),
    view='file_metadata:file_metadata_driver_list'
)



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_group_view, permission_user_view
from .querysets import get_user_queryset

search_model_group = SearchModel(
    app_label='auth', label=_(message='Group'), model_name='Group',
    permission=permission_group_view,
    serializer_path='mayan.apps.user_management.serializers.GroupSerializer'
)

search_model_group.add_model_field(
    field='name', label=_(message='Name')
)

user_app, user_model = settings.AUTH_USER_MODEL.split('.')

search_model_user = SearchModel(
    app_label=user_app, label=_(message='User'), model_name=user_model,
    permission=permission_user_view, queryset=get_user_queryset,
    serializer_path='mayan.apps.user_management.serializers.UserSerializer'
)

search_model_user.add_model_field(
    field='first_name', label=_(message='First name')
)
search_model_user.add_model_field(
    field='email', label=_(message='Email')
)
search_model_user.add_model_field(
    field='groups__name', label=_(message='Groups')
)
search_model_user.add_model_field(
    field='last_name', label=_(message='Last name')
)
search_model_user.add_model_field(
    field='username', label=_(message='Username')
)



from mayan.apps.icons.icons import Icon

icon_asset_create = Icon(driver_name='fontawesome', symbol='save')
icon_asset_delete_multiple = Icon(driver_name='fontawesome', symbol='times')
icon_asset_delete_single = icon_asset_delete_multiple
icon_asset_detail = Icon(driver_name='fontawesome', symbol='save')
icon_asset_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_asset_list = Icon(driver_name='fontawesome', symbol='save')

icon_transformations = Icon(driver_name='fontawesome', symbol='crop')

icon_transformation_create = Icon(driver_name='fontawesome', symbol='plus')
icon_transformation_delete = Icon(driver_name='fontawesome', symbol='times')
icon_transformation_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_transformation_list = icon_transformations
icon_transformation_select = Icon(driver_name='fontawesome', symbol='plus')

icon_layer_decorations = Icon(
    driver_name='fontawesome', symbol='paint-roller'
)
icon_layer_transformation = Icon(
    driver_name='fontawesome', symbol='crop'
)



import email
import logging

from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import flatten_map
from mayan.apps.file_metadata.classes import FileMetadataDriver
from mayan.apps.storage.literals import MIME_TYPE_EML

logger = logging.getLogger(__name__)


class EMLDriver(FileMetadataDriver):
    description = _(
        message='Extracts information from emails saved in .eml files.'
    )
    internal_name = 'eml'
    label = _(message='EML Python driver')
    mime_type_list = list(MIME_TYPE_EML)

    def _get_message_metadata(self, message, metadata_map):
        metadata_map.update(
            message.items()
        )
        filename = message.get_filename()
        if filename:
            metadata_map['filename'] = filename

        if message.is_multipart():
            metadata_map['parts'] = {}

            iterator_parts = message.iter_parts()

            for index, part in enumerate(iterable=iterator_parts):
                metadata_map['parts'][index] = {}
                metadata_map_part = metadata_map['parts'][index]

                self._get_message_metadata(
                    message=part, metadata_map=metadata_map_part
                )

    def _process(self, document_file):
        result = {}

        with document_file.open() as file_object:
            message = email.message_from_binary_file(
                fp=file_object, policy=email.policy.default
            )
            metadata_map = {}
            self._get_message_metadata(
                message=message, metadata_map=metadata_map
            )
            flatten_map(dictionary=metadata_map, result=result)

        return result



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import Tag
from .permissions import permission_tag_attach, permission_tag_remove


class TagSerializer(serializers.HyperlinkedModelSerializer):
    documents_url = serializers.HyperlinkedIdentityField(
        label=_(message='Documents URL'), lookup_url_kwarg='tag_id',
        view_name='rest_api:tag-document-list'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'tag_id',
                'view_name': 'rest_api:tag-detail'
            }
        }
        fields = (
            'color', 'documents_url', 'id', 'label', 'url'
        )
        model = Tag


class DocumentTagAttachSerializer(serializers.Serializer):
    tag = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the tag to add to the document.'
        ), label=_(message='Tag ID'), source_model=Tag,
        source_permission=permission_tag_attach
    )


class DocumentTagRemoveSerializer(serializers.Serializer):
    tag = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the tag to remove from the document.'
        ), label=_(message='Tag ID'), source_model=Tag,
        source_permission=permission_tag_remove
    )



import logging

from django.apps import apps
from django.db import models

from mayan.apps.converter.settings import setting_image_generation_timeout
from mayan.apps.lock_manager.backends.base import LockingBackend

from .classes import OCRBackendBase
from .events import (
    event_ocr_document_version_content_deleted,
    event_ocr_document_version_finished
)
from .exceptions import OCRError
from .literals import ERROR_LOG_DOMAIN_NAME

logger = logging.getLogger(name=__name__)


class DocumentVersionPageOCRContentManager(models.Manager):
    def delete_content_for(self, document_version, user=None):
        self.filter(document_version_page__document_version=document_version).delete()

        event_ocr_document_version_content_deleted.commit(
            actor=user, action_object=document_version.document,
            target=document_version
        )

    def do_ocr_finished(self, document_version, user):
        event_ocr_document_version_finished.commit(
            action_object=document_version.document, actor=user,
            target=document_version
        )

    def process_document_version_page(
        self, document_version_page, user=None
    ):
        logger.debug(
            'Processing page: %d of document version: %s',
            document_version_page.page_number,
            document_version_page.document_version
        )

        DocumentVersionPageOCRContent = apps.get_model(
            app_label='ocr', model_name='DocumentVersionPageOCRContent'
        )

        lock_name = document_version_page.get_lock_name(user=user)

        try:
            document_version_page_lock = LockingBackend.get_backend().acquire_lock(
                name=lock_name,
                timeout=setting_image_generation_timeout.value * 2
            )
        except Exception as exception:
            logger.error(
                'Error attempting to lock document version page: %d; %s',
                document_version_page.pk, exception, exc_info=True
            )
            raise
        else:
            try:
                cache_filename = document_version_page.generate_image(
                    _acquire_lock=False, user=user
                )

                with document_version_page.cache_partition.get_file(filename=cache_filename).open() as file_object:
                    try:
                        ocr_content = OCRBackendBase.get_instance().execute(
                            file_object=file_object,
                            language=document_version_page.document_version.document.language
                        )
                    except OCRError as exception:
                        document_version_page.error_log.create(
                            domain_name=ERROR_LOG_DOMAIN_NAME,
                            text=str(exception)
                        )
                    else:
                        DocumentVersionPageOCRContent.objects.update_or_create(
                            document_version_page=document_version_page,
                            defaults={
                                'content': ocr_content
                            }
                        )
                        queryset_error_logs = document_version_page.error_log.filter(
                            domain_name=ERROR_LOG_DOMAIN_NAME
                        )
                        queryset_error_logs.delete()
            except Exception as exception:
                logger.error(
                    'OCR error for document version page: %d; %s',
                    document_version_page.pk, exception, exc_info=True
                )
                raise
            else:
                logger.debug(
                    'Finished processing page: %d of document version: %s',
                    document_version_page.page_number,
                    document_version_page.document_version
                )
            finally:
                document_version_page_lock.release()


class DocumentTypeSettingsManager(models.Manager):
    def get_by_natural_key(self, document_type_natural_key):
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        try:
            document_type = DocumentType.objects.get_by_natural_key(
                document_type_natural_key
            )
        except DocumentType.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(document_type__pk=document_type.pk)



from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    SingleObjectCreateView, SingleObjectDeleteView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms.workflow_template_transition_forms import WorkflowTransitionForm
from ..icons import (
    icon_workflow_template_transition,
    icon_workflow_template_transition_create,
    icon_workflow_template_transition_delete,
    icon_workflow_template_transition_edit,
    icon_workflow_template_transition_list
)
from ..links import link_workflow_template_transition_create
from ..models import Workflow, WorkflowTransition
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)


class WorkflowTemplateTransitionCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_id'
    form_class = WorkflowTransitionForm
    view_icon = icon_workflow_template_transition_create

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                message='Create transitions for workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['workflow'] = self.external_object
        return kwargs

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.transitions.all()

    def get_success_url(self):
        return reverse(
            kwargs={
                'workflow_template_id': self.kwargs[
                    'workflow_template_id'
                ]
            }, viewname='document_states:workflow_template_transition_list'
        )


class WorkflowTemplateTransitionDeleteView(SingleObjectDeleteView):
    model = WorkflowTransition
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'navigation_object_list': ('object', 'workflow'),
            'title': _(
                message='Delete workflow transition: %s?'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_success_url(self):
        return reverse(
            kwargs={'workflow_template_id': self.object.workflow.pk},
            viewname='document_states:workflow_template_transition_list'
        )


class WorkflowTemplateTransitionEditView(SingleObjectEditView):
    form_class = WorkflowTransitionForm
    model = WorkflowTransition
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.object,
            'title': _(
                message='Edit workflow transition: %s'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['workflow'] = self.object.workflow
        return kwargs

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_success_url(self):
        return reverse(
            kwargs={'workflow_template_id': self.object.workflow.pk},
            viewname='document_states:workflow_template_transition_list'
        )


class WorkflowTemplateTransitionListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_id'
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_template_transition_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_template_transition,
            'no_results_main_link': link_workflow_template_transition_create.resolve(
                context=RequestContext(
                    dict_={'workflow': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                message='Create a transition and use it to move a workflow '
                'from one state to another.'
            ),
            'no_results_title': _(
                message='This workflow doesn\'t have any transitions'
            ),
            'object': self.external_object,
            'title': _(
                message='Transitions of workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.transitions.all()



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_object
)
from mayan.apps.navigation.source_columns import SourceColumn

from .links import (
    link_user_confirm_view_property_delete_multiple,
    link_user_confirm_view_property_delete_single,
    link_user_confirm_view_property_list, link_user_view_modes
)


class ViewsApp(MayanAppConfig):
    app_namespace = 'views'
    app_url = 'views'
    has_tests = True
    name = 'mayan.apps.views'
    verbose_name = _(message='Views')

    def ready(self):
        super().ready()

        UserConfirmView = self.get_model(model_name='UserConfirmView')
        UserViewMode = self.get_model(model_name='UserViewMode')
        User = get_user_model()

        SourceColumn(
            attribute='get_app_config_verbose_name', include_label=True,
            source=UserConfirmView
        )
        SourceColumn(
            attribute='name', is_identifier=True, is_sortable=True,
            source=UserConfirmView
        )
        SourceColumn(
            attribute='remember', include_label=True, is_sortable=True,
            source=UserConfirmView
        )

        SourceColumn(
            attribute='get_app_config_verbose_name', include_label=True,
            source=UserViewMode
        )
        SourceColumn(
            attribute='name', is_identifier=True, is_sortable=True,
            source=UserViewMode
        )
        SourceColumn(
            attribute='get_value_display', include_label=True,
            source=UserViewMode
        )

        menu_list_facet.bind_links(
            links=(
                link_user_confirm_view_property_list, link_user_view_modes,
            ), sources=(User,)
        )

        menu_multi_item.bind_links(
            links=(link_user_confirm_view_property_delete_multiple,),
            sources=(UserConfirmView,)
        )

        menu_object.bind_links(
            links=(
                link_user_confirm_view_property_delete_single,
            ), sources=(UserConfirmView,)
        )



from mayan.apps.common.utils import comma_splitter
from mayan.apps.templating.template_backends import Template


class MetadataTypeBusinessLogicMixin:
    def get_default_value(self):
        template = Template(template_string=self.default)
        return template.render()

    def get_lookup_values(self):
        template = Template(
            context_entry_name_list=('groups', 'users'),
            template_string=self.lookup
        )

        template_result = template.render()

        return comma_splitter(string=template_result)

    def get_required_for(self, document_type):
        """
        Determine if the metadata type is required for the specified document
        type.
        """
        queryset = document_type.metadata.filter(
            required=True, metadata_type=self
        )

        return queryset.exists()



from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentAPIViewMixin
)
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)
from mayan.apps.rest_api import generics

from .models import DuplicateBackendEntry
from .serializers import DuplicateTargetDocumentSerializer


class APIDuplicatedDocumentListView(generics.ListAPIView):
    """
    get: Return a list of duplicated documents.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer
    source_queryset = DuplicateBackendEntry.objects.get_duplicated_documents()


class APIDocumentDuplicateListView(
    ParentObjectDocumentAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of the selected document's duplicates.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DuplicateTargetDocumentSerializer

    def get_source_queryset(self):
        return DuplicateBackendEntry.objects.get_duplicates_of(
            document=self.get_document(
                permission=permission_document_view
            )
        )



from django.core import management

from mayan.apps.document_indexing.models.index_instance_models import (
    IndexInstance
)

from ..base import MountCommandMixin


class Command(MountCommandMixin, management.BaseCommand):
    help = 'Mount an index as a filesystem.'
    node_identifier_argument = 'slug'
    node_identifier_help_text = 'Index slug'
    node_text_attribute = 'value'

    def factory_func_document_container_node(self, *args, **options):
        index_instance = IndexInstance.objects.get(
            **{
                self.node_identifier_argument: options[
                    self.node_identifier_argument
                ]
            }
        )

        def func_document_container_node():
            return index_instance.get_root()

        return func_document_container_node



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_comments', '0004_auto_20150920_0202'),
    ]

    operations = [
        migrations.RenameField(
            model_name='comment',
            old_name='comment',
            new_name='text',
        ),
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_ORGANIZATIONS_INSTALLATION_URL,
    DEFAULT_ORGANIZATIONS_URL_BASE_PATH
)
from .setting_validators import validation_fuction_check_path_format

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Organizations'), name='organizations'
)

setting_organization_installation_url = setting_namespace.do_setting_add(
    default=DEFAULT_ORGANIZATIONS_INSTALLATION_URL,
    global_name='ORGANIZATIONS_INSTALLATION_URL',
    help_text=_(
        message='Fully qualified domain including scheme, and port number if using '
        'a custom one. Do not include the path. Use the '
        'ORGANIZATIONS_URL_BASE_PATH setting to specify the path. '
        'Example: https://www.example.com:8080'
    )
)
setting_organization_url_base_path = setting_namespace.do_setting_add(
    default=DEFAULT_ORGANIZATIONS_URL_BASE_PATH,
    global_name='ORGANIZATIONS_URL_BASE_PATH',
    help_text=_(
        message='Base URL path to use for all views. Used when hosting using a path '
        'that is not the root path of the web server. The path value '
        'must not include a leading or trailing slash. '
        'Example: mayan-installations/home'
    ), validation_function=validation_fuction_check_path_format
)



import functools

from django.core.exceptions import ImproperlyConfigured
from django.template.utils import EngineHandler


class Template:
    @classmethod
    @functools.cache
    def get_backend(cls):
        engine_handler = EngineHandler(
            templates=(
                {
                    'BACKEND': 'django.template.backends.django.DjangoTemplates',
                    'OPTIONS': {
                        'builtins': [
                            'mayan.apps.templating.templatetags.templating_datetime_tags',
                            'mayan.apps.templating.templatetags.templating_json_tags',
                            'mayan.apps.templating.templatetags.templating_math_tags',
                            'mayan.apps.templating.templatetags.templating_regex_tags',
                            'mayan.apps.templating.templatetags.templating_tags',
                            'mayan.apps.templating.templatetags.templating_yaml_tags'
                        ]
                    }
                },
            )
        )
        return engine_handler['django']

    def __init__(self, template_string, context_entry_name_list=None):
        self._template_backend = Template.get_backend()
        self.context_entry_name_list = context_entry_name_list or ()

        self._template = self._template_backend.from_string(
            template_code=template_string
        )

    def render(self, context=None):
        final_context = {}

        entries_context = TemplateContextEntry.get_as_context(
            entry_name_list=self.context_entry_name_list
        )
        final_context.update(entries_context)

        if context:
            final_context.update(context)

        return self._template.render(context=final_context)


class TemplateContextEntry:
    _registry = {}
    _registry_always_available = []

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_as_context(cls, entry_name_list):
        result = {}

        for entry_name in cls._registry_always_available:
            entry = cls.get(name=entry_name)
            result[entry.name] = entry.get_value()

        for entry_name in entry_name_list:
            entry = cls.get(name=entry_name)
            result[entry.name] = entry.get_value()

        return result

    @classmethod
    def get_as_help_text(cls, entry_name_list):
        result = []

        for entry_name in cls._registry_always_available:
            entry = cls.get(name=entry_name)
            help_text = entry.get_help_text()
            result.append(help_text)

        for entry_name in entry_name_list:
            entry = cls.get(name=entry_name)
            help_text = entry.get_help_text()
            result.append(help_text)

        return ', '.join(result)

    def __init__(self, description, name, value, always_available=False):
        self.always_available = always_available
        self.description = description
        self.name = name
        self.value = value

        if name in self.__class__._registry:
            raise ImproperlyConfigured(
                '{} entry `{}` already registered. Check spelling or change '
                'the name'.format(self.__class__, self.name)
            )

        self.__class__._registry[self.name] = self

        if always_available:
            self.__class__._registry_always_available.append(self.name)

    def get_help_text(self):
        return '{{{{ {name} }}}} - {description}'.format(
            name=self.name, description=self.description
        )

    def get_value(self):
        try:
            return self.value()
        except TypeError:
            return self.value



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import DEFAULT_FORMS_SHOW_DROPZONE_SUBMIT_BUTTON

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Forms'), name='forms'
)

setting_show_dropzone_submit_button = setting_namespace.do_setting_add(
    default=DEFAULT_FORMS_SHOW_DROPZONE_SUBMIT_BUTTON,
    global_name='FORMS_SHOW_DROPZONE_SUBMIT_BUTTON', help_text=_(
        message='Display a submit button and wait for the submit button to be '
        'pressed before processing up the upload.'
    )
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE,
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND,
    DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS
)
from .setting_callbacks import callback_update_signature_capture_cache_size

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Signature captures'), name='signature_captures',
)


setting_signature_capture_cache_maximum_size = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_MAXIMUM_SIZE',
    help_text=_(
        message='The threshold at which the SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND '
        'will start deleting the oldest signature capture cache files. '
        'Specify the size in bytes.'
    ), post_edit_function=callback_update_signature_capture_cache_size
)
setting_signature_capture_cache_storage_backend = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND',
    help_text=_(
        message='Path to the Storage subclass to use when storing the cached '
        'signature capture files.'
    )
)
setting_signature_capture_cache_storage_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        message='Arguments to pass to the SIGNATURE_CAPTURES_SIGNATURE_CAPTURE_CACHE_STORAGE_BACKEND.'
    )
)



from django.db import migrations
import django.db.models.deletion

import mptt.fields


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0018_auto_20200402_0647')
    ]

    operations = [
        migrations.AlterField(
            model_name='indexinstancenode',
            name='parent',
            field=mptt.fields.TreeForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='children',
                to='document_indexing.IndexInstanceNode'
            )
        ),
        migrations.AlterField(
            model_name='indextemplatenode',
            name='parent',
            field=mptt.fields.TreeForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='children',
                to='document_indexing.IndexTemplateNode'
            )
        )
    ]



from mayan.apps.forms import form_fields, form_widgets, forms

from .classes import Dependency


class DependenciesLicensesForm(forms.Form):
    text = form_fields.CharField(
        label='',
        widget=form_widgets.TextAreaDiv(
            attrs={
                'class': 'full-height scrollable',
                'data-height-difference': 270
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        text_legal_list = []

        for dependency in Dependency.get_all():
            text_legal = dependency.get_legal_text()
            if text_legal:
                text_legal_list.append(
                    '-' * len(
                        dependency.get_label()
                    )
                )
                text_legal_list.append(
                    dependency.get_label().strip()
                )
                text_legal_list.append(
                    '-' * len(
                        dependency.get_label()
                    )
                )

                # Implement word wrapping at 79 columns.
                for line in text_legal.split('\n'):
                    line_length = 0
                    new_line = []

                    for word in line.strip().split():
                        if line_length + len(word) > 79:
                            text_legal_list.append(
                                ' '.join(new_line)
                            )
                            new_line = [word]
                            line_length = 0
                        else:
                            new_line.append(word)
                            line_length += len(word)

                    text_legal_list.append(
                        ' '.join(new_line)
                    )

                text_legal_list.append('\n')

        self.fields['text'].initial = '\n'.join(text_legal_list)



from django.urls import path, re_path

from .views.authentication_views import (
    MayanLoginView, MayanLogoutView, MayanPasswordChangeDoneView,
    MayanPasswordChangeView, MultiFactorAuthenticationView,
    UserSetPasswordView
)
from .views.impersonation_views import (
    UserImpersonateEndView, UserImpersonateFormStartView,
    UserImpersonateStartView
)
from .views.password_reset_views import (
    MayanPasswordResetCompleteView, MayanPasswordResetConfirmView,
    MayanPasswordResetDoneView, MayanPasswordResetView
)

urlpatterns_authenticattion = [
    re_path(
        route=r'^login/$', name='login_view',
        view=MayanLoginView.as_view()
    ),
    re_path(
        route=r'^login/multi_factor_authentication/$',
        name='multi_factor_authentication_view',
        view=MultiFactorAuthenticationView.as_view()
    ),
    re_path(
        route=r'^logout/$', view=MayanLogoutView.as_view(),
        name='logout_view'
    )
]

urlpatterns_password_reset = [
    re_path(
        route=r'^password/reset/complete/$',
        name='password_reset_complete_view',
        view=MayanPasswordResetCompleteView.as_view()
    ),
    path(
        'password/reset/confirm/<uidb64>/<token>/',
        name='password_reset_confirm_view',
        view=MayanPasswordResetConfirmView.as_view()
    ),
    re_path(
        route=r'^password/reset/done/$',
        name='password_reset_done_view',
        view=MayanPasswordResetDoneView.as_view()
    ),
    re_path(
        route=r'^password/reset/$', name='password_reset_view',
        view=MayanPasswordResetView.as_view()
    )
]

urlpatterns_password_set = [
    re_path(
        route=r'^password/change/done/$', name='password_change_done',
        view=MayanPasswordChangeDoneView.as_view()
    ),
    re_path(
        route=r'^password/change/$', name='password_change_view',
        view=MayanPasswordChangeView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/set_password/$',
        name='user_set_password', view=UserSetPasswordView.as_view()
    ),
    re_path(
        route=r'^users/multiple/set_password/$',
        name='user_multiple_set_password',
        view=UserSetPasswordView.as_view()
    )
]

urlpatterns_user_impersonation = [
    re_path(
        route=r'^impersonate/end/$', name='user_impersonate_end',
        view=UserImpersonateEndView.as_view()
    ),
    re_path(
        route=r'^impersonate/start/$', name='user_impersonate_form_start',
        view=UserImpersonateFormStartView.as_view()
    ),
    re_path(
        route=r'^impersonate/(?P<user_id>\d+)/start/$',
        name='user_impersonate_start',
        view=UserImpersonateStartView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_authenticattion)
urlpatterns.extend(urlpatterns_password_reset)
urlpatterns.extend(urlpatterns_password_set)
urlpatterns.extend(urlpatterns_user_impersonation)



from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse

from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.user_management.serializers import UserSerializer

from .models import SavedResultset


class DummySearchResultModelSerializer(serializers.Serializer):
    """
    Empty serializer for Swagger views.
    """


class SavedResultsetSerializer(serializers.HyperlinkedModelSerializer):
    results_url = serializers.HyperlinkedIdentityField(
        label=_(message='Results URL'), lookup_url_kwarg='saved_resultset_id',
        view_name='rest_api:saved_resultset-result-list'
    )
    user = UserSerializer(
        label=_(message='User'), read_only=True
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'saved_resultset_id',
                'view_name': 'rest_api:saved_resultset-detail'
            }
        }
        fields = (
            'app_label', 'id', 'model_name', 'result_count', 'results_url',
            'search_explainer_text', 'time_to_live', 'timestamp', 'url',
            'user'
        )
        model = SavedResultset
        read_only_fields = fields

    def create(self, validated_data):
        saved_resultset, queryset = self.context['view'].do_search_execute(
            store_resultset=True
        )

        return saved_resultset

    def validate(self, data):
        if not self.context['view'].request.user.is_authenticated:
            raise ValidationError(
                detail=_(message='User must be authenticated.')
            )

        return data


class SearchFieldSerializer(serializers.Serializer):
    field_class_label = serializers.CharField(
        label=_(message='Field class'), read_only=True
    )
    field_name = serializers.CharField(
        label=_(message='Field name'), read_only=True
    )
    get_help_text = serializers.CharField(
        label=_(message='Description'), read_only=True
    )
    get_search_field_class_label = serializers.CharField(
        label=_(message='Search field class'), read_only=True
    )
    label = serializers.CharField(
        label=_(message='Label'), read_only=True
    )


class SearchModelSerializer(serializers.Serializer):
    app_label = serializers.CharField(
        label=_(message='App label'), read_only=True
    )
    model_name = serializers.CharField(
        label=_(message='Model name'), read_only=True
    )
    pk = serializers.CharField(
        label=_(message='Primary key'), read_only=True
    )
    search_fields = SearchFieldSerializer(
        label=_(message='Search fields'), many=True, read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL')
    )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:searchmodel-detail', kwargs={
                'search_model_pk': instance.full_name
            }, request=self.context['request'], format=self.context['format']
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import JavaScriptDependency

JavaScriptDependency(
    label=_(message='Fancybox'), module=__name__, name='@fancyapps/fancybox',
    version_string='=3.2.5'
)
JavaScriptDependency(
    label=_(message='jQuery'), module=__name__, name='jquery',
    version_string='=3.7.1'
)
JavaScriptDependency(
    label=_(message='JQuery Form'), module=__name__, name='jquery-form',
    version_string='=4.3.0'
)
JavaScriptDependency(
    label=_(message='JQuery Match Height'), module=__name__,
    name='jquery-match-height', version_string='=0.7.2'
)
JavaScriptDependency(
    label=_(message='Select 2'), module=__name__, name='select2',
    version_string='=4.0.13'
)
JavaScriptDependency(
    label=_(message='Toastr'), module=__name__, name='toastr',
    version_string='=2.1.4'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import forms

from .models import AccessControlList


class ACLCreateForm(forms.FilteredSelectionForm, forms.ModelForm):
    class Meta:
        field_name = 'role'
        fields = ('role',)
        label = _(message='Role')
        model = AccessControlList
        widget_attributes = {'class': 'select2'}



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .utils import get_language_option_list, get_timezone_option_list

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Locales'), name='locales'
)

setting_user_language_default = setting_namespace.do_setting_add(
    choices=get_language_option_list(), default=settings.LANGUAGE_CODE,
    global_name='LOCALES_USER_DEFAULT_LANGUAGE', help_text=_(
        message='Default profile language for new user accounts. Must be in '
        'ISO639-3 format.'
    )
)
setting_user_timezone_default = setting_namespace.do_setting_add(
    choices=get_timezone_option_list(), default=settings.TIME_ZONE,
    global_name='LOCALES_USER_DEFAULT_TIMEZONE', help_text=_(
        message='Default profile timezone for new user accounts.'
    )
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_duplicates = CeleryQueue(
    label=_(message='Duplicates'), name='duplicates', worker=worker_c
)
queue_duplicates_slow = CeleryQueue(
    label=_(message='Duplicates slow'), name='duplicates_slow', worker=worker_c
)

queue_duplicates.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_clean_empty_lists',
    label=_(message='Clean empty duplicate lists')
)
queue_duplicates.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_scan_for',
    label=_(message='Scan document duplicates')
)

queue_duplicates_slow.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_scan_all',
    label=_(message='Duplicated document scan')
)



import json

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import F

from .search_models import SearchModel
from .settings import setting_saved_resultset_time_to_live_increment


class SavedResultsetBusinessLogicModelMixin:
    def get_content_type(self):
        content_type = ContentType.objects.get(
            app_label=self.app_label, model=self.model_name
        )

        return content_type

    def get_model_class(self):
        content_type = self.get_content_type()

        return content_type.model_class()

    def get_result_queryset(self):
        SavedResultset = apps.get_model(
            app_label='dynamic_search', model_name='SavedResultset'
        )

        SavedResultset.objects.update(
            time_to_live=F('time_to_live') + setting_saved_resultset_time_to_live_increment.value
        )

        content_type = self.get_content_type()

        values_list = self.entries.values_list('object_id', flat=True)

        queryset = content_type.get_all_objects_for_this_type()
        queryset = queryset.filter(pk__in=values_list)

        return queryset

    def get_search_model(self):
        model_class = self.get_model_class()

        return SearchModel.get_for_model(instance=model_class)

    def get_search_query_dict(self):
        obj = json.loads(s=self.search_query_text)

        return obj



from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.views import (
    ViewSingleObjectDynamicFormModelBackendCreate,
    ViewSingleObjectDynamicFormModelBackendEdit
)
from mayan.apps.views.generics import (
    FormView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..classes import WorkflowAction
from ..forms.workflow_template_state_forms import (
    WorkflowTemplateStateActionDynamicForm,
    WorkflowTemplateStateActionSelectionForm, WorkflowTemplateStateForm
)
from ..icons import (
    icon_workflow_template_state, icon_workflow_template_state_action,
    icon_workflow_template_state_action_create,
    icon_workflow_template_state_action_delete,
    icon_workflow_template_state_action_edit,
    icon_workflow_template_state_action_list,
    icon_workflow_template_state_action_selection,
    icon_workflow_template_state_create, icon_workflow_template_state_delete,
    icon_workflow_template_state_edit, icon_workflow_template_state_list
)
from ..links import (
    link_workflow_template_state_action_selection,
    link_workflow_template_state_create
)
from ..models import Workflow, WorkflowState, WorkflowStateAction
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)


class WorkflowTemplateStateActionCreateView(
    ExternalObjectViewMixin, ViewSingleObjectDynamicFormModelBackendCreate
):
    backend_class = WorkflowAction
    external_object_class = WorkflowState
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_state_id'
    form_class = WorkflowTemplateStateActionDynamicForm
    view_icon = icon_workflow_template_state_action_create

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.external_object,
            'title': _(
                message='Create a "%(backend_label)s" workflow action for: '
                '%(workflow_state)s'
            ) % {
                'backend_label': self.get_backend_class().label,
                'workflow_state': self.external_object
            },
            'workflow': self.external_object.workflow
        }

    def get_form_extra_kwargs(self):
        return {
            'request': self.request,
            'user': self.request.user
        }

    def get_form_schema_extra_kwargs(self):
        return {'workflow_template_state': self.external_object}

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'backend_path': self.kwargs['backend_path'],
            'state': self.external_object
        }

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_state_id': self.external_object.pk
            },
            viewname='document_states:workflow_template_state_action_list'
        )


class WorkflowTemplateStateActionDeleteView(SingleObjectDeleteView):
    model = WorkflowStateAction
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_action_id'
    view_icon = icon_workflow_template_state_action_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_state', 'workflow'
            ),
            'object': self.object,
            'title': _(
                message='Delete workflow state action: %s'
            ) % self.object,
            'workflow': self.object.state.workflow,
            'workflow_state': self.object.state
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_state_id': self.object.state.pk
            },
            viewname='document_states:workflow_template_state_action_list'
        )


class WorkflowTemplateStateActionEditView(
    ViewSingleObjectDynamicFormModelBackendEdit
):
    form_class = WorkflowTemplateStateActionDynamicForm
    model = WorkflowStateAction
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_action_id'
    view_icon = icon_workflow_template_state_action_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_state', 'workflow'
            ),
            'object': self.object,
            'title': _(
                message='Edit workflow state action: %s'
            ) % self.object,
            'workflow': self.object.state.workflow,
            'workflow_state': self.object.state
        }

    def get_form_extra_kwargs(self):
        return {
            'request': self.request,
            'user': self.request.user
        }

    def get_form_schema_extra_kwargs(self):
        return {'workflow_template_state': self.object.state}

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'workflow_template_state_id': self.object.state.pk
            },
            viewname='document_states:workflow_template_state_action_list'
        )


class WorkflowTemplateStateActionListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = WorkflowState
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_state_id'
    view_icon = icon_workflow_template_state_action_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': ('object', 'workflow'),
            'no_results_icon': icon_workflow_template_state_action,
            'no_results_main_link': link_workflow_template_state_action_selection.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                message='Workflow state actions are macros that get executed '
                'when documents enters or leaves the state in which they '
                'reside.'
            ),
            'no_results_title': _(
                message='There are no actions for this workflow state'
            ),
            'object': self.external_object,
            'title': _(
                message='Actions for workflow state: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_source_queryset(self):
        return self.external_object.actions.all()


class WorkflowTemplateStateActionSelectionView(
    ExternalObjectViewMixin, FormView
):
    external_object_class = WorkflowState
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_state_id'
    form_class = WorkflowTemplateStateActionSelectionForm
    view_icon = icon_workflow_template_state_action_selection

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow'
            ),
            'object': self.external_object,
            'title': _(
                message='New workflow state action selection for: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def form_valid(self, form):
        klass = form.cleaned_data['klass']
        return HttpResponseRedirect(
            redirect_to=reverse(
                kwargs={
                    'workflow_template_state_id': self.external_object.pk,
                    'backend_path': klass
                },
                viewname='document_states:workflow_template_state_action_create'
            )
        )


class WorkflowTemplateStateCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_id'
    form_class = WorkflowTemplateStateForm
    view_icon = icon_workflow_template_state_create

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                message='Create states for workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.states.all()

    def get_success_url(self):
        return reverse(
            kwargs={
                'workflow_template_id': self.kwargs['workflow_template_id']
            },
            viewname='document_states:workflow_template_state_list'
        )


class WorkflowTemplateStateDeleteView(SingleObjectDeleteView):
    model = WorkflowState
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_id'
    view_icon = icon_workflow_template_state_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.object,
            'title': _(
                message='Delete workflow state: %s?'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_success_url(self):
        return reverse(
            kwargs={
                'workflow_template_id': self.object.workflow.pk
            },
            viewname='document_states:workflow_template_state_list'
        )


class WorkflowTemplateStateEditView(SingleObjectEditView):
    form_class = WorkflowTemplateStateForm
    model = WorkflowState
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_state_id'
    view_icon = icon_workflow_template_state_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.object,
            'title': _(
                message='Edit workflow state: %s'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_success_url(self):
        return reverse(
            kwargs={'workflow_template_id': self.object.workflow.pk},
            viewname='document_states:workflow_template_state_list'
        )


class WorkflowTemplateStateListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_id'
    view_icon = icon_workflow_template_state_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_template_state,
            'no_results_main_link': link_workflow_template_state_create.resolve(
                context=RequestContext(
                    dict_={'workflow': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                message='Create states and link them using transitions.'
            ),
            'no_results_title': _(
                message='This workflow doesn\'t have any states'
            ),
            'object': self.external_object,
            'title': _(
                message='States of workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.states.all()



from django.core.exceptions import PermissionDenied

from rest_framework.permissions import BasePermission

from mayan.apps.acls.models import AccessControlList
from mayan.apps.permissions.classes import Permission


class MayanPermission(BasePermission):
    def get_mayan_object_permission(self, request, view):
        try:
            method_get_mayan_object_permission = getattr(
                view, 'get_mayan_object_permission'
            )
        except AttributeError:
            mayan_object_permission_map = getattr(
                view, 'mayan_object_permission_map', {}
            )

            permission = mayan_object_permission_map.get(
                request.method, None
            )
            return permission
        else:
            return method_get_mayan_object_permission()

    def get_mayan_view_permission(self, request, view):
        try:
            method_get_mayan_view_permission_map = getattr(
                view, 'get_mayan_view_permission_map'
            )
        except AttributeError:
            mayan_view_permission_map = getattr(
                view, 'mayan_view_permission_map', {}
            )

            permission = mayan_view_permission_map.get(
                request.method, None
            )

            return permission
        else:
            return method_get_mayan_view_permission_map()

    def has_object_permission(self, request, view, obj):
        permission = self.get_mayan_object_permission(
            request=request, view=view
        )
        user = request.user

        if permission:
            if user.is_authenticated:
                try:
                    AccessControlList.objects.check_access(
                        obj=obj, permission=permission, user=user
                    )
                except PermissionDenied:
                    return False
                else:
                    return True
            else:
                return False
        else:
            return True

    def has_permission(self, request, view):
        permission = self.get_mayan_view_permission(
            request=request, view=view
        )
        user = request.user

        if permission:
            if user.is_authenticated:
                try:
                    Permission.check_user_permission(
                        permission=permission, user=user
                    )
                except PermissionDenied:
                    return False
                else:
                    return True
            else:
                return False
        else:
            return True



from mayan.apps.documents.permissions import permission_document_type_view
from mayan.apps.documents.serializers.document_type_serializers import (
    DocumentTypeSerializer
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..models.index_template_models import IndexTemplate
from ..permissions import (
    permission_index_template_create, permission_index_template_delete,
    permission_index_template_edit, permission_index_template_rebuild,
    permission_index_template_view
)
from ..serializers.index_template_serializers import (
    DocumentTypeAddSerializer, DocumentTypeRemoveSerializer,
    IndexTemplateSerializer
)
from ..tasks import task_index_template_rebuild

from .index_template_api_view_mixins import APIIndexTemplateNodeViewMixin


class APIIndexTemplateListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the defined indexes template.
    post: Create a new index template.
    """
    mayan_object_permission_map = {'GET': permission_index_template_view}
    mayan_view_permission_map = {'POST': permission_index_template_create}
    serializer_class = IndexTemplateSerializer
    source_queryset = IndexTemplate.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIIndexTemplateDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected index template.
    get: Returns the details of the selected index template.
    patch: Partially edit an index template.
    put: Edit an index template.
    """
    lookup_url_kwarg = 'index_template_id'
    mayan_object_permission_map = {
        'DELETE': permission_index_template_delete,
        'GET': permission_index_template_view,
        'PATCH': permission_index_template_edit,
        'PUT': permission_index_template_edit
    }
    serializer_class = IndexTemplateSerializer
    source_queryset = IndexTemplate.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIIndexTemplateDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of the document types associated with this index template.
    """
    external_object_class = IndexTemplate
    external_object_pk_url_kwarg = 'index_template_id'
    mayan_external_object_permission_map = {
        'GET': permission_index_template_view
    }
    mayan_object_permission_map = {'GET': permission_document_type_view}
    serializer_class = DocumentTypeSerializer

    def get_source_queryset(self):
        return self.get_external_object().document_types.all()


class APIIndexTemplateDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to an index template.
    """
    lookup_url_kwarg = 'index_template_id'
    mayan_object_permission_map = {'POST': permission_index_template_edit}
    serializer_class = DocumentTypeAddSerializer
    source_queryset = IndexTemplate.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_add(
            queryset=[document_type], user=self.request.user
        )


class APIIndexTemplateDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from an index template.
    """
    lookup_url_kwarg = 'index_template_id'
    mayan_object_permission_map = {'POST': permission_index_template_edit}
    serializer_class = DocumentTypeRemoveSerializer
    source_queryset = IndexTemplate.objects.all()

    def object_action(self, obj, request, serializer):
        document_type = serializer.validated_data['document_type']
        obj.document_types_remove(
            queryset=[document_type], user=self.request.user
        )


class APIIndexTemplateNodeListView(
    APIIndexTemplateNodeViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the template nodes for the selected index.
    post: Create a new index template node.
    """
    def get_source_queryset(self):
        return self.get_index_template().index_template_root_node.get_children()


class APIIndexTemplateNodeDetailView(
    APIIndexTemplateNodeViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected index template node.
    get: Returns the details of the selected index template node.
    patch: Partially edit an index template node.
    put: Edit an index template node.
    """
    lookup_url_kwarg = 'index_template_node_id'

    def get_source_queryset(self):
        return self.get_index_template().index_template_nodes.all()


class APIIndexTemplateRebuildView(generics.ObjectActionAPIView):
    """
    post: Rebuild the selected index template.
    """
    lookup_url_kwarg = 'index_template_id'
    mayan_object_permission_map = {
        'POST': permission_index_template_rebuild
    }
    source_queryset = IndexTemplate.objects.all()

    def object_action(self, obj, request, serializer):
        task_index_template_rebuild.apply_async(
            kwargs={
                'index_id': obj.pk
            }
        )


class APIIndexTemplateResetView(generics.ObjectActionAPIView):
    """
    post: Reset the selected index template.
    """
    lookup_url_kwarg = 'index_template_id'
    mayan_object_permission_map = {
        'POST': permission_index_template_rebuild
    }
    source_queryset = IndexTemplate.objects.all()

    def object_action(self, obj, request, serializer):
        obj.reset()



from django.db import migrations, models
import mayan.apps.common.validators


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0016_storeddriver_exists')
    ]

    operations = [
        migrations.AddField(
            field=models.TextField(
                blank=True, help_text='Enter the arguments for the drive '
                'for the specific document type as a YAML dictionary. ie: '
                '{"degrees": 180}', validators=[
                    mayan.apps.common.validators.YAMLValidator()
                ], verbose_name='Arguments'
            ),
            model_name='documenttypedriverconfiguration',
            name='arguments'
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_fields, form_widgets, forms

from .fields import ModelTemplateField


class TemplateSandboxForm(forms.Form):
    result = form_fields.CharField(
        help_text=_(message='Resulting text from the evaluated template.'),
        label=_(message='Result'), required=False, widget=form_widgets.Textarea(
            attrs={
                'class': 'appearance-overscroll-contain appearance-resize-vertical appearance-resize-persist',
                'readonly': 'readonly', 'rows': 3
            }
        )
    )

    def __init__(self, *args, **kwargs):
        self.model = kwargs.pop('model')
        self.model_variable = kwargs.pop('model_variable')
        super().__init__(*args, **kwargs)
        self.fields['template'] = ModelTemplateField(
            initial_help_text=_(
                message='The template string to be evaluated.'
            ), label=_(message='Template'), model=self.model,
            model_variable=self.model_variable, required=True
        )
        self.order_fields(
            field_order=('template', 'result')
        )



from django.db.models import CharField, Value
from django.db.models.expressions import OuterRef, Subquery
from django.db.models.functions import Concat
from django.utils import timezone
from django.utils.module_loading import import_string

from actstream.models import Action

from .settings import (
    setting_event_prune_backend, setting_event_prune_backend_arguments
)


class EventLogPruneBackend:
    @classmethod
    def get_backend_class(cls):
        return import_string(dotted_path=setting_event_prune_backend.value)

    @classmethod
    def get_backend_instance(cls):
        backend_class = cls.get_backend_class()

        return backend_class(**setting_event_prune_backend_arguments.value)

    def _execute(self):
        return NotImplementedError

    def execute(self):
        self._execute()


class EventLogPruneBackendLatest(EventLogPruneBackend):
    """
    Keep the last N events in the entire log.
    """

    def __init__(self, number):
        self.number = number

    def _execute(self):
        queryset_remain = Action.objects.order_by('-timestamp').values_list('pk')[:self.number]
        queryset_delete = Action.objects.exclude(pk__in=queryset_remain)
        queryset_delete.delete()


class EventLogPruneBackendLatestPerObject(EventLogPruneBackend):
    """
    Keep the last N events for each target.
    """

    def __init__(self, number):
        self.number = number

    def _execute(self):
        queryset_annotated = Action.objects.values(
            'target_content_type', 'target_object_id', 'verb'
        ).annotate(
            ct_fk_combination=Concat(
                'target_content_type', Value('-'), 'target_object_id',
                output_field=CharField()
            )
        )

        queryset_lastest_per_object = queryset_annotated.filter(
            pk__in=Subquery(
                queryset_annotated.filter(
                    ct_fk_combination=OuterRef('ct_fk_combination')
                ).order_by('-timestamp').values('pk')[:self.number]
            )
        )

        queryset_delete = Action.objects.exclude(
            pk__in=queryset_lastest_per_object.values_list('pk')
        )

        queryset_delete.delete()


class EventLogPruneBackendLatestPerObjectEventType(EventLogPruneBackend):
    """
    Keep the last N events for each target per event type. Ensures
    at least the most recent event for each type is retained.
    """

    def __init__(self, number):
        self.number = number

    def _execute(self):
        queryset_annotated = Action.objects.values(
            'target_content_type', 'target_object_id', 'verb'
        ).annotate(
            ct_fk_verb_combination=Concat(
                'target_content_type', Value('-'), 'target_object_id',
                Value('-'), 'verb', output_field=CharField()
            )
        )

        queryset_lastest_per_object = queryset_annotated.filter(
            pk__in=Subquery(
                queryset_annotated.filter(
                    ct_fk_verb_combination=OuterRef('ct_fk_verb_combination')
                ).order_by('-timestamp').values('pk')[:self.number]
            )
        )

        queryset_delete = Action.objects.exclude(
            pk__in=queryset_lastest_per_object.values_list('pk')
        )

        queryset_delete.delete()


class EventLogPruneBackendOlderThanDays(EventLogPruneBackend):
    """
    Delete events older than N days.
    """

    def __init__(self, days):
        self.days = days

    def _execute(self):
        cutoff_datetime = timezone.now() - timezone.timedelta(days=self.days)
        queryset = Action.objects.filter(timestamp__lt=cutoff_datetime)
        queryset.delete()



import hashlib

from django.conf import settings
from django.core import serializers
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view

from ..literals import (
    ERROR_LOG_DOMAIN_NAME, GRAPHVIZ_COLOR_STATE_FILL_FINAL,
    GRAPHVIZ_COLOR_STATE_FILL_INITIAL, GRAPHVIZ_COLOR_STATE_FONT_COLOR_FINAL,
    GRAPHVIZ_ID_STATE, GRAPHVIZ_SHAPE_CIRCLE, GRAPHVIZ_SHAPE_DOUBLECIRCLE,
    GRAPHVIZ_STYLE_FILLED, WORKFLOW_ACTION_ON_ENTRY, WORKFLOW_ACTION_ON_EXIT
)


class WorkflowStateBusinessLogicMixin:
    def do_active_set(self, log_entry=None, workflow_instance=None):
        # TODO: Update one initial entry log patch is merged.
        # Break pattern by allowing `workflow_instance` and log_entry=None
        # until initial entry log patch is merged.
        queryset = self.entry_actions.filter(enabled=True)
        self.do_queryset_actions_execute(
            log_entry=log_entry, queryset=queryset,
            workflow_instance=workflow_instance
        )

    def do_active_unset(self, log_entry):
        queryset = self.exit_actions.filter(enabled=True)
        self.do_queryset_actions_execute(
            log_entry=log_entry, queryset=queryset
        )

    def do_queryset_actions_execute(
        self, queryset, log_entry=None, workflow_instance=None
    ):
        # TODO: Update one initial entry log patch is merged.
        # Break pattern by allowing `workflow_instance` and log_entry=None
        # until initial entry log patch is merged.
        if log_entry:
            workflow_instance = log_entry.workflow_instance

        for action in queryset:
            context = workflow_instance.get_context()
            context.update(
                {'action': action, 'log_entry': log_entry}
            )

            try:
                action.execute(
                    context=context, workflow_instance=workflow_instance
                )
            except Exception as exception:
                queryset_error_logs = workflow_instance.document.error_log
                queryset_error_logs.create(
                    domain_name=ERROR_LOG_DOMAIN_NAME,
                    text='{}; {}'.format(
                        exception.__class__.__name__, exception
                    )
                )

                if settings.DEBUG or settings.TESTING:
                    raise

                break
            else:
                queryset_error_logs = workflow_instance.document.error_log.filter(
                    domain_name=ERROR_LOG_DOMAIN_NAME
                )
                queryset_error_logs.delete()

    def do_diagram_generate(self, diagram):
        fill_color = ''
        font_color = ''
        style = ''

        is_edge_state = self.initial or not self.destination_transitions.exists() or not self.origin_transitions.exists()

        if is_edge_state:
            shape = GRAPHVIZ_SHAPE_DOUBLECIRCLE
        else:
            shape = GRAPHVIZ_SHAPE_CIRCLE

        if self.final:
            fill_color = GRAPHVIZ_COLOR_STATE_FILL_FINAL
            font_color = GRAPHVIZ_COLOR_STATE_FONT_COLOR_FINAL
            style = GRAPHVIZ_STYLE_FILLED
        elif self.initial:
            fill_color = GRAPHVIZ_COLOR_STATE_FILL_INITIAL
            style = GRAPHVIZ_STYLE_FILLED

        node_kwargs = {
            'fillcolor': fill_color,
            'fontcolor': font_color,
            'label': self.label,
            'name': self.get_graph_id(),
            'shape': shape,
            'style': style
        }
        diagram.node(**node_kwargs)

        for action in self.actions.all():
            action.do_diagram_generate(diagram=diagram)

        for escalation in self.escalations.all():
            escalation.do_diagram_generate(diagram=diagram)

    @property
    def entry_actions(self):
        return self.actions.filter(when=WORKFLOW_ACTION_ON_ENTRY)

    @property
    def exit_actions(self):
        return self.actions.filter(when=WORKFLOW_ACTION_ON_EXIT)

    def get_actions_display(self):
        field_list = [
            str(field) for field in self.actions.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_actions_display.short_description = _(message='Actions')

    def get_escalations_display(self):
        field_list = [
            str(field) for field in self.escalations.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_escalations_display.short_description = _(message='Escalations')

    def get_graph_id(self):
        return '{}{}'.format(GRAPHVIZ_ID_STATE, self.pk)

    def get_hash(self):
        result = hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        )

        for action in self.actions.all():
            result.update(
                action.get_hash().encode()
            )

        for escalation in self.escalations.all():
            result.update(
                escalation.get_hash().encode()
            )

        return result.hexdigest()


class WorkflowStateRuntimeProxyBusinessLogicMixin:
    def get_documents(self, permission=None, user=None):
        """
        Provide a queryset of the documents. The queryset is optionally
        filtered by access.
        """
        if self.workflow.ignore_completed:
            queryset = Document.valid.none()
        else:
            queryset = Document.valid.filter(workflows__state_active=self)

        if permission and user:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=user
            )

        return queryset

    def get_document_count(self, user):
        """
        Return the numeric count of documents at this workflow state.
        The count is filtered by access.
        """
        return self.get_documents(
            permission=permission_document_view, user=user
        ).count()



from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import DynamicFormModelBackend

from .events import event_credential_used


class CredentialBackend(DynamicFormModelBackend):
    _backend_app_label = 'credentials'
    _backend_model_name = 'StoredCredential'
    _loader_module_name = 'credential_backends'

    form_fieldsets = (
        (
            _(message='General'), {
                'fields': ('label', 'internal_name')
            }
        ),
    )

    def get_credential(self, action_object=None, user=None):
        model_instance = self.get_model_instance()

        backend_data = model_instance.get_backend_data()
        event_credential_used.commit(
            action_object=action_object, actor=user, target=model_instance
        )
        return backend_data


# Null backend must be defined here to avoid automatic import.
class CredentialBackendNull(CredentialBackend):
    label = _(message='Null backend')



from django.urls import re_path

from .views import (
    CacheDetailView, CacheListView, CachePartitionDetailView,
    CachePartitionPurgeView, CachePurgeView
)

urlpatterns_apps = [
    re_path(
        route=r'^apps/(?P<app_label>[-\w]+)/models/(?P<model_name>[-\w]+)/objects/(?P<object_id>\d+)/cache_partitions/purge/$',
        name='cache_partitions_purge', view=CachePartitionPurgeView.as_view()
    )
]

urlpatterns_caches = [
    re_path(
        route=r'^caches/$', name='cache_list', view=CacheListView.as_view()
    ),
    re_path(
        route=r'^caches/(?P<cache_id>\d+)/detail/$', name='cache_detail',
        view=CacheDetailView.as_view()
    ),
    re_path(
        route=r'^caches/(?P<cache_id>\d+)/purge/$', name='cache_purge',
        view=CachePurgeView.as_view()
    ),
    re_path(
        route=r'^caches/multiple/purge/$', name='cache_multiple_purge',
        view=CachePurgeView.as_view()
    ),
    re_path(
        route=r'^caches/(?P<cache_partition_id>\d+)/$',
        name='cache_partition_detail',
        view=CachePartitionDetailView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_apps)
urlpatterns.extend(urlpatterns_caches)



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from google.cloud import storage

from .base import DownloadBackend


class DownloadBackendGoogleCloudStorageSignedURL(DownloadBackend):
    DEFAULT_EXPIRATION = 3600  # 1 hour

    def __init__(
        self, bucket_name, stored_credential_internal_name, expiration=None
    ):
        self.bucket_name = bucket_name
        self.stored_credential_internal_name = stored_credential_internal_name

        if not expiration:
            expiration = self.DEFAULT_EXPIRATION

        self.expiration = expiration

    def get_download_file_object(self, obj):
        StoredCredential = apps.get_model(
            app_label='credentials', model_name='StoredCredential'
        )

        stored_credential = StoredCredential.objects.get(
            internal_name=self.stored_credential_internal_name
        )

        stored_credential_data = stored_credential.get_backend_data()

        storage_client = storage.Client.from_service_account_info(
            info=stored_credential_data
        )

        bucket = storage_client.bucket(bucket_name=self.bucket_name)

        filename = self.get_stored_filename(obj=obj)
        blob = bucket.blob(blob_name=filename)

        return blob.generate_signed_url(
            expiration=self.expiration, method='GET', version='v4'
        )

    def get_download_filename(self, obj):
        super_filename = super().get_download_filename(obj=obj)
        return 'signed-url-for-{}'.format(super_filename)

    def get_stored_filename(self, obj):
        raise ImproperlyConfigured(
            'Class `{}` must provide a `get_stored_filename` method '
            'that returns the storage filename.'.format(
                self.__class__.__name__
            )
        )



import warnings

from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import menu_list_facet, menu_return, menu_tools
from mayan.apps.navigation.source_columns import SourceColumn

from .classes import ModelProperty, ModelWrapper
from .literals import MESSAGE_SQLITE_WARNING
from .links import link_model_property_list, link_property_model_list
from .patches import patch_Migration
from .utils import check_for_sqlite
from .warnings import DatabaseWarning


class DatabasesApp(MayanAppConfig):
    app_namespace = 'databases'
    app_url = 'databases'
    has_tests = True
    name = 'mayan.apps.databases'
    verbose_name = _(message='Databases')

    def ready(self):
        super().ready()

        if check_for_sqlite():
            warnings.warn(
                category=DatabaseWarning,
                message=str(MESSAGE_SQLITE_WARNING)
            )

        patch_Migration()

        # ModelProperty

        SourceColumn(
            attribute='get_label', include_label=True, is_identifier=True,
            source=ModelProperty
        )
        SourceColumn(
            attribute='get_description', include_label=True,
            source=ModelProperty
        )
        SourceColumn(
            attribute='get_name', include_label=True, source=ModelProperty
        )

        # ModelWrapper

        SourceColumn(
            attribute='get_name_full', include_label=True, is_identifier=True,
            source=ModelWrapper
        )
        SourceColumn(
            attribute='get_app_label', include_label=True,
            source=ModelWrapper
        )
        SourceColumn(
            attribute='get_label', include_label=True,
            source=ModelWrapper
        )

        # menus

        menu_tools.bind_links(
            links=(link_property_model_list,)
        )

        menu_list_facet.bind_links(
            links=(link_model_property_list,), sources=(ModelWrapper,)
        )

        menu_return.bind_links(
            links=(link_property_model_list,), sources=(ModelWrapper,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SingleObjectListView
from mayan.apps.views.view_mixins import ContentTypeViewMixin

from .classes import ModelProperty, ModelWrapper
from .icons import icon_model_property_list, icon_property_model_list


class ModelPropertyListView(ContentTypeViewMixin, SingleObjectListView):
    content_type_url_kw_args = {
        'app_label': 'app_label',
        'model_name': 'model_name'
    }
    view_icon = icon_model_property_list

    def get_extra_context(self):
        model = self.get_model()
        verbose_name = model._meta.verbose_name

        model_wrapper = ModelWrapper(model=model)

        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': self.view_icon,
            'no_results_title': _(message='No properties available'),
            'object': model_wrapper,
            'title': _(message='Properties for model: %s') % verbose_name
        }

    def get_model(self):
        content_type = self.get_content_type()
        return content_type.model_class()

    def get_source_queryset(self):
        model = self.get_model()

        result = list(
            ModelProperty.get_for(model=model)
        )
        return result


class PropertyModelListView(SingleObjectListView):
    extra_context = {
        'hide_link': True,
        'hide_object': True,
        'title': _(message='Models with registered properties')
    }
    view_icon = icon_property_model_list

    def get_source_queryset(self):
        result = ModelWrapper.all()
        return result



from django.urls import re_path

from mayan.apps.views.generics import SimpleView

urlpatterns_error_pages = [
    re_path(
        route=r'^errors/403/$', name='error_403', view=SimpleView.as_view(
            template_name='403.html'
        )
    ),
    re_path(
        route=r'^errors/404/$', name='error_404', view=SimpleView.as_view(
            template_name='404.html'
        )
    ),
    re_path(
        route=r'^errors/500/$', name='error_500', view=SimpleView.as_view(
            template_name='500.html'
        )
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_error_pages)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Metadata'), name='metadata'
)

event_document_metadata_added = namespace.add_event_type(
    label=_(message='Document metadata added'), name='document_metadata_added'
)
event_document_metadata_edited = namespace.add_event_type(
    label=_(message='Document metadata edited'), name='document_metadata_edited'
)
event_document_metadata_removed = namespace.add_event_type(
    label=_(message='Document metadata removed'), name='document_metadata_removed'
)
event_metadata_type_created = namespace.add_event_type(
    label=_(message='Metadata type created'), name='metadata_type_created'
)
event_metadata_type_edited = namespace.add_event_type(
    label=_(message='Metadata type edited'), name='metadata_type_edited'
)
event_metadata_type_relationship_updated = namespace.add_event_type(
    label=_(message='Metadata type relationship updated'),
    name='metadata_type_relationship'
)



from django.urls import re_path

from .api_views import (
    APIErrorLogPartitionEntryDetailView, APIErrorLogPartitionEntryListView
)
from .views import (
    GlobalErrorLogEntryList, ObjectErrorLogEntryDeleteView,
    ObjectErrorLogEntryListClearView, ObjectErrorLogEntryListView
)

urlpatterns_objects = [
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/errors/$',
        name='object_error_log_entry_list', view=ObjectErrorLogEntryListView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/errors/clear/$',
        name='object_error_log_entry_list_clear',
        view=ObjectErrorLogEntryListClearView.as_view()
    ),
    re_path(
        route=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/errors/(?P<error_log_partition_entry_id>\d+)/delete/$',
        name='object_error_log_entry_delete',
        view=ObjectErrorLogEntryDeleteView.as_view()
    )
]

urlpatterns_error_logs = [
    re_path(
        route=r'^error_logs/partitions/entries/$',
        name='global_error_log_partition_entry_list',
        view=GlobalErrorLogEntryList.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_objects)
urlpatterns.extend(urlpatterns_error_logs)

api_urls = [
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/errors/$',
        name='errorlogpartitionentry-list',
        view=APIErrorLogPartitionEntryListView.as_view()
    ),
    re_path(
        route=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/errors/(?P<error_log_partition_entry_id>\d+)/$',
        name='errorlogpartitionentry-detail',
        view=APIErrorLogPartitionEntryDetailView.as_view()
    )
]



from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from rest_framework.relations import HyperlinkedIdentityField

from mayan.apps.common.utils import resolve_attribute

from . import serializers
from .field_mixins import (
    AutoHelpTextLabelFieldMixin, FilteredRelatedFieldMixin
)


class FilteredPrimaryKeyRelatedField(
    AutoHelpTextLabelFieldMixin, FilteredRelatedFieldMixin,
    serializers.PrimaryKeyRelatedField
):
    """
    PrimaryKeyRelatedField that allows runtime queryset filtering by ACL.
    """


class FilteredSimplePrimaryKeyRelatedField(
    AutoHelpTextLabelFieldMixin, FilteredRelatedFieldMixin,
    serializers.RelatedField
):
    """
    PrimaryKeyRelatedField that allows runtime queryset filtering by ACL
    and that only stores the primary key.
    """
    default_error_messages = {
        'does_not_exist': _(
            message='Invalid pk "{pk_value}" - object does not exist.'
        ),
        'incorrect_type': _(
            message='Incorrect type. Expected pk value, received {data_type}.'
        ),
        'required': _(message='This field is required.')
    }

    def __init__(self, **kwargs):
        self.pk_field = kwargs.pop('pk_field', 'pk')
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        kwargs = {
            self.pk_field: data
        }

        queryset = self.get_queryset()
        try:
            return getattr(
                queryset.get(**kwargs), self.pk_field
            )
        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=data)
        except (TypeError, ValueError):
            self.fail(
                'incorrect_type', data_type=type(data).__name__
            )

    def to_representation(self, value):
        return getattr(value, self.pk_field)


class MultiKwargHyperlinkedIdentityField(HyperlinkedIdentityField):
    def __init__(self, *args, **kwargs):
        self.view_kwargs = kwargs.pop(
            'view_kwargs', []
        )
        super().__init__(*args, **kwargs)

    def get_url(self, obj, view_name, request, format):
        """
        Extends HyperlinkedRelatedField to allow passing more than one view
        keyword argument.
        ----
        Given an object, return the URL that hyperlinks to the object.

        May raise a `NoReverseMatch` if the `view_name` and `lookup_field`
        attributes are not configured to correctly match the URL conf.
        """
        # Unsaved objects will not yet have a valid URL.
        if hasattr(obj, 'pk') and obj.pk in (None, ''):
            return None

        kwargs = {}
        for entry in self.view_kwargs:
            kwargs[
                entry['lookup_url_kwarg']
            ] = resolve_attribute(
                attribute=entry['lookup_field'], obj=obj
            )

        return self.reverse(
            format=format, kwargs=kwargs, request=request, viewname=view_name
        )



import logging

from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..models.document_models import Document
from ..models.document_type_models import DocumentType
from ..permissions import (
    permission_document_type_create, permission_document_type_delete,
    permission_document_type_edit, permission_document_type_view,
    permission_document_view
)
from ..serializers.document_serializers import DocumentSerializer
from ..serializers.document_type_serializers import (
    DocumentTypeQuickLabelSerializer, DocumentTypeSerializer
)

from .api_view_mixins import ParentObjectDocumentTypeAPIViewMixin

logger = logging.getLogger(name=__name__)


class APIDocumentTypeDocumentListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the documents of the selected document types.
    """
    external_object_class = DocumentType
    external_object_pk_url_kwarg = 'document_type_id'
    mayan_external_object_permission_map = {
        'GET': permission_document_type_view
    }
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer

    def get_source_queryset(self):
        external_object = self.get_external_object()
        queryset_documents_all = external_object.documents.values('id')
        return Document.valid.filter(pk__in=queryset_documents_all)


class APIDocumentTypeListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the document types.
    post: Create a new document type.
    """
    mayan_object_permission_map = {'GET': permission_document_type_view}
    mayan_view_permission_map = {'POST': permission_document_type_create}
    serializer_class = DocumentTypeSerializer
    source_queryset = DocumentType.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIDocumentTypeDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected document type.
    get: Return the details of the selected document type.
    patch: Edit the properties of the selected document type.
    put: Edit the properties of the selected document type.
    """
    lookup_url_kwarg = 'document_type_id'
    mayan_object_permission_map = {
        'DELETE': permission_document_type_delete,
        'GET': permission_document_type_view,
        'PATCH': permission_document_type_edit,
        'PUT': permission_document_type_edit
    }
    serializer_class = DocumentTypeSerializer
    source_queryset = DocumentType.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIDocumentTypeQuickLabelDetailView(
    ParentObjectDocumentTypeAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected quick label.
    get: Return the details of the selected quick label.
    patch: Edit the properties of the selected quick label.
    put: Edit the properties of the selected quick label.
    """
    lookup_url_kwarg = 'document_type_quick_label_id'
    mayan_object_permission_map = {
        'DELETE': permission_document_type_edit,
        'GET': permission_document_type_view,
        'PATCH': permission_document_type_edit,
        'PUT': permission_document_type_edit
    }
    serializer_class = DocumentTypeQuickLabelSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.get_document_type().filenames.all()


class APIDocumentTypeQuickLabelListView(
    ParentObjectDocumentTypeAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the document type quick labels.
    post: Create a new document type quick label.
    """
    serializer_class = DocumentTypeQuickLabelSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'document_type': self.get_document_type(
                permission=permission_document_type_edit
            )
        }

    def get_source_queryset(self):
        # This method is only called during GET, therefore filter only by
        # the view permission.
        return self.get_document_type(
            permission=permission_document_type_view
        ).filenames.all()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0017_auto_20211226_1036')
    ]

    operations = [
        migrations.AlterField(
            model_name='documentmetadata',
            name='value',
            field=models.TextField(
                blank=True, default='', help_text='The actual value '
                'stored in the metadata type field for the document.',
                verbose_name='Value'
            ), preserve_default=False
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Comments'), name='comments'
)

permission_document_comment_create = namespace.add_permission(
    label=_(message='Create new comments'), name='comment_create'
)
permission_document_comment_delete = namespace.add_permission(
    label=_(message='Delete comments'), name='comment_delete'
)
permission_document_comment_edit = namespace.add_permission(
    label=_(message='Edit comments'), name='comment_edit'
)
permission_document_comment_view = namespace.add_permission(
    label=_(message='View comments'), name='comment_view'
)



import yaml

from django.utils.safestring import SafeString

try:
    from yaml import CSafeLoader as SafeLoader, CSafeDumper as SafeDumper
except ImportError:
    from yaml import SafeLoader, SafeDumper


def yaml_dump(**kwargs):
    defaults = {'Dumper': SafeDumper}
    defaults.update(kwargs)

    return yaml.dump(**defaults)


def yaml_load(**kwargs):
    stream = kwargs['stream']

    if isinstance(stream, SafeString):
        # Convert SafeStr to str.
        stream = stream.strip()
        kwargs['stream'] = stream

    defaults = {'Loader': SafeLoader}
    defaults.update(kwargs)

    return yaml.load(**defaults)



from django.utils.translation import gettext_lazy as _

from mayan.apps.icons.icons import Icon
from mayan.apps.source_compressed.source_backends.mixins import (
    SourceBackendMixinCompressed
)
from mayan.apps.source_interactive.source_backends.mixins import (
    SourceBackendMixinInteractive
)
from mayan.apps.source_stored_files.source_backends.storage_backend_source_mixins import (
    SourceBackendMixinStoredFileLocationStorageBackend
)
from mayan.apps.source_stored_files.source_backends.stored_file_source_mixins import (
    SourceBackendMixinStoredFileInteractive
)
from mayan.apps.sources.source_backends.base import SourceBackend
from mayan.apps.sources.source_backends.mixins import (
    SourceBackendMixinRegularExpression
)


class SourceBackendStagingStorage(
    SourceBackendMixinCompressed, SourceBackendMixinInteractive,
    SourceBackendMixinRegularExpression,
    SourceBackendMixinStoredFileInteractive,
    SourceBackendMixinStoredFileLocationStorageBackend, SourceBackend
):
    icon = Icon(driver_name='fontawesome', symbol='file')
    label = _(message='Staging storage')
